# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Test data functions

Functions creating test data: curves, images, ...
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import dataclasses

import numpy as np

from codraft.core.computation import fit
from codraft.core.io.image import imread_scor
from codraft.core.io.signal import read_signal
from codraft.core.model.base import ResultShape, ShapeTypes
from codraft.core.model.image import create_image
from codraft.core.model.signal import create_signal
from codraft.utils.tests import get_test_fnames


def add_gaussian_noise(
    data: np.ndarray, mu: float = 0.0, sigma: float = 0.1, seed: int = 1
) -> None:
    """Add Gaussian (Normal-law) random noise to data"""
    rng = np.random.default_rng(seed)
    data += rng.normal(mu, sigma, size=data.shape)


def create_1d_gaussian(
    size,
    amp: float = 50.0,
    sigma: float = 2.0,
    x0: float = 0.0,
    y0: float = 0.0,
    noise_mu: float = 0.0,
    noise_sigma: float = 0.0,
    seed: int = 1,
):
    """Create Gaussian curve data (optionally noised, if noise_sigma != 0.0)"""
    x = np.linspace(-10, 10, size)
    y = fit.GaussianModel.func(x, amp, sigma, x0, y0)
    if noise_sigma:
        add_gaussian_noise(y, mu=noise_mu, sigma=noise_sigma, seed=seed)
    return x, y


def create_test_2d_data(size, dtype):
    """Creating 2D test data"""
    x, y = np.meshgrid(np.linspace(0, 10, size), np.linspace(0, 10, size))
    raw_data = 0.5 * (np.sin(x) + np.cos(y)) + 0.5
    dmin = np.iinfo(dtype).min * 0.95
    dmax = np.iinfo(dtype).max * 0.95
    return np.array(raw_data * (dmax - dmin) + dmin, dtype=dtype)


def create_2d_steps_data(size, width, dtype):
    """Creating 2D steps data for testing purpose"""
    data = np.zeros((size, size), dtype=dtype)
    value = 1
    for col in range(0, size - width + 1, width):
        data[:, col : col + width] = value
        value *= 10
    data2 = np.zeros_like(data)
    value = 1
    for row in range(0, size - width + 1, width):
        data2[row : row + width, :] = value
        value *= 10
    data += data2
    return data


def create_2d_random(size, dtype, level=0.1, seed=1):
    """Creating 2D Uniform-law random image"""
    rng = np.random.default_rng(seed)
    amp = np.iinfo(dtype).max * level
    return np.array(rng.random((size, size)) * amp, dtype=dtype)


def create_2d_gaussian(size, dtype, x0=0, y0=0, mu=0.0, sigma=2.0, amp=None):
    """Creating 2D Gaussian (-10 <= x <= 10 and -10 <= y <= 10)"""
    xydata = np.linspace(-10, 10, size)
    x, y = np.meshgrid(xydata, xydata)
    if amp is None:
        amp = np.iinfo(dtype).max * 0.5
    return np.array(
        amp
        * np.exp(
            -((np.sqrt((x - x0) ** 2 + (y - y0) ** 2) - mu) ** 2) / (2.0 * sigma**2)
        ),
        dtype=dtype,
    )


def get_laser_spot_data():
    """Return a list of NumPy arrays containing images which are relevant for
    testing laser spot image processing features"""
    znoise = create_2d_random(2000, np.uint16)
    zgauss = create_2d_gaussian(2000, np.uint16, x0=2.0, y0=-3.0)
    return [zgauss + znoise] + [
        imread_scor(fname) for fname in get_test_fnames("*.scor-data")
    ]


@dataclasses.dataclass
class PeakDataParam:
    """Peak data test image parameters"""

    size: int = 2000
    n_points: int = 4
    sigma_gauss2d: float = 0.06
    amp_gauss2d: int = 1900
    mu_noise: int = 845
    sigma_noise: int = 25
    dx0: float = 0.0
    dy0: float = 0.0
    att: float = 1.0


def get_peak2d_data(p: PeakDataParam = None, seed=None, multi=False):
    """Return a list of NumPy arrays containing images which are relevant for
    testing 2D peak detection or similar image processing features"""
    if p is None:
        p = PeakDataParam()
    delta = 0.1
    rng = np.random.default_rng(seed)
    coords = (rng.random((p.n_points, 2)) - 0.5) * 10 * (1 - delta)
    data = rng.normal(p.mu_noise, p.sigma_noise, size=(p.size, p.size))
    multi_nb = 2 if multi else 1
    for x0, y0 in coords:
        for idx in range(multi_nb):
            if idx != 0:
                p.dx0 = 0.08 + rng.random() * 0.08
                p.dy0 = 0.08 + rng.random() * 0.08
                p.att = 0.2 + rng.random() * 0.8
            data += create_2d_gaussian(
                p.size,
                np.uint16,
                x0=x0 + p.dx0,
                y0=y0 + p.dy0,
                sigma=p.sigma_gauss2d,
                amp=p.amp_gauss2d / multi_nb * p.att,
            )
    return data


def __get_default_data_size(size=None):
    """Return default data size"""
    return 500 if size is None else size


def create_test_signal1(size=None, title=None):
    """Create test signal (Paracetamol molecule spectrum)"""
    obj = read_signal(get_test_fnames("paracetamol.txt")[0])
    if title is not None:
        obj.title = title
    size = __get_default_data_size(size)
    if size is not None:
        x0, y0 = obj.xydata
        x1 = np.linspace(x0[0], x0[-1], size)
        y1 = np.interp(x1, x0, y0)
        obj.set_xydata(x1, y1)
    return obj


def create_test_signal2(size=None, title=None, noised=False):
    """Create test signal (Gaussian curve, optionally noised)"""
    size = __get_default_data_size(size)
    default_title = "Gaussian curve" + (" with noise" if noised else "")
    obj = create_signal(default_title if title is None else title)
    x, y = create_1d_gaussian(size=size, noise_sigma=5.0 if noised else 0.0)
    obj.set_xydata(x, y)
    return obj


def __get_default_size_dtype(size=None, dtype=None):
    """Return default shape and dtype"""
    size = 2000 if size is None else size
    dtype = np.uint16 if dtype is None else dtype
    return size, dtype


def create_test_image1(size=None, dtype=None, title=None):
    """Create test image (sin(x)+cos(y))"""
    size, dtype = __get_default_size_dtype(size, dtype)
    title = "sin(x)+cos(y)" if title is None else title
    return create_image(title, create_test_2d_data(size, dtype=dtype))


def create_test_image2(size=None, dtype=None, with_annotations=True, title=None):
    """Create test image (2D noisy gaussian)"""
    size, dtype = __get_default_size_dtype(size, dtype)
    title = "2D Gaussian" if title is None else title
    zgauss = create_2d_gaussian(size, dtype=dtype, x0=2.0, y0=3.0)
    znoise = create_2d_random(size, dtype)
    data = zgauss + znoise
    image = create_image(title, data)
    if with_annotations:
        with open(get_test_fnames("annotations.json")[0], mode="rb") as fdesc:
            image.annotations = fdesc.read().decode()
    return image


def create_test_image3(size=None, dtype=None, title=None):
    """Create test image (multiple 2D-gaussian peaks)"""
    size, dtype = __get_default_size_dtype(size, dtype)
    data = (
        create_2d_gaussian(size, dtype, x0=0.5, y0=3.0)
        + create_2d_gaussian(size, dtype, x0=-1.0, y0=-1.0, sigma=1.0)
        + create_2d_gaussian(size, dtype, x0=7.0, y0=8.0)
    )
    return create_image("Multiple 2D-gaussian peaks" if title is None else title, data)


def create_resultshapes():
    """Create test result shapes (core.model.base.ResultShape test objects)"""
    return (
        ResultShape(
            ShapeTypes.CIRCLE,
            [[0, 100, 100, 400, 400], [0, 150, 150, 350, 350]],
            "circle",
        ),
        ResultShape(ShapeTypes.RECTANGLE, [0, 300, 200, 700, 700], "rectangle"),
        ResultShape(ShapeTypes.SEGMENT, [0, 50, 250, 400, 400], "segment"),
        ResultShape(ShapeTypes.POINT, [[0, 500, 500], [0, 15, 400]], "point"),
    )


def create_image_with_annotations():
    """Create test image with annotations"""
    data = create_2d_gaussian(600, np.uint16, x0=2.0, y0=3.0)
    image = create_image("Test image with metadata", data)
    with open(get_test_fnames("annotations.json")[0], mode="rb") as fdesc:
        image.annotations = fdesc.read().decode()
    return image



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Configuration utilities
"""

import os
import os.path as osp

from guidata.userconfig import NoDefault, UserConfig, get_config_dir

CONF = UserConfig({})


class Configuration:
    """Configuration file"""

    @classmethod
    def initialize(cls, name, version, load):
        """Initialize configuration"""
        CONF.set_application(name, version, load=load)

    @classmethod
    def reset(cls):
        """Reset configuration"""
        global CONF  # pylint: disable=global-statement
        CONF.cleanup()  # Remove configuration file
        CONF = UserConfig({})


class Section:
    """Configuration section"""

    @classmethod
    def set_name(cls, section):
        """Set section name"""
        cls._name = section

    @classmethod
    def get_name(cls):
        """Return section name"""
        return cls._name


class Option:
    """Configuration option handler"""

    def __init__(self):
        self.section = None
        self.option = None

    def get(self, default=NoDefault):
        """Get configuration option value"""
        return CONF.get(self.section, self.option, default)

    def set(self, value):
        """Set configuration option value"""
        CONF.set(self.section, self.option, value)

    def reset(self):
        """Reset configuration option"""
        CONF.remove_option(self.section, self.option)


class ConfigPathOption(Option):
    """Configuration file path configuration option handler"""

    def get(self, default=NoDefault):
        """Get configuration file path from configuration"""
        if default is NoDefault:
            default = ""
        fname = super().get(default)
        if osp.basename(fname) != fname:
            raise ValueError("Invalid configuration file name {fname}")
        return osp.join(get_config_dir(), osp.basename(fname))


class WorkingDirOption(Option):
    """Working directory configuration option handler"""

    def get(self, default=NoDefault):
        """Get working directory from configuration"""
        if default is NoDefault:
            default = ""
        path = super().get(default)
        if osp.isdir(path):
            return path
        return ""

    def set(self, value):
        """Set working directory in configuration"""
        if not osp.isdir(value):
            value = osp.dirname(value)
            if not osp.isdir(value):
                raise FileNotFoundError(f"Invalid working directory name {value}")
        os.chdir(value)
        super().set(value)


class SectionMeta(type):
    """Configuration metaclass"""

    def __new__(cls, name, bases, dct):
        optlist = []
        for attrname, obj in list(dct.items()):
            if isinstance(obj, Option):
                obj.option = attrname
                optlist.append(obj)
        dct["_options"] = optlist
        return type.__new__(cls, name, bases, dct)


class ConfMeta(type):
    """Configuration metaclass"""

    def __new__(cls, name, bases, dct):
        for attrname, obj in list(dct.items()):
            if isinstance(obj, Section):
                obj.set_name(attrname)
                for option in obj._options:
                    option.section = attrname
        return type.__new__(cls, name, bases, dct)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Unit test scenario Image 01

Testing the following:
  - Add image object at startup
  - Create image (random type)
  - Sum two images
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

from codraft.core.gui.main import CodraFTMainWindow
from codraft.core.gui.processor.image import (
    ContourShapeParam,
    LogP1Param,
    PeakDetectionParam,
    ResizeParam,
    RotateParam,
    ZCalibrateParam,
)
from codraft.core.model.base import UniformRandomParam
from codraft.core.model.image import ImageTypes, create_image, new_image_param
from codraft.tests import codraft_app_context
from codraft.tests.data import PeakDataParam, create_test_image1, get_peak2d_data
from codraft.tests.newobject_unit import iterate_image_creation
from codraft.tests.scenario_sig_app import test_common_operations

SHOW = True  # Show test in GUI-based test launcher


def test_image_features(win: CodraFTMainWindow, data_size: int = 150) -> None:
    """Testing signal features"""
    win.switch_to_image_panel()
    panel = win.imagepanel

    for image in iterate_image_creation(data_size, non_zero=True):
        panel.add_object(create_test_image1(data_size))
        panel.add_object(image)
        test_common_operations(panel)
        panel.remove_all_objects()

    ima1 = create_test_image1(data_size)
    panel.add_object(ima1)

    # Add new image based on i0
    panel.objlist.set_current_row(0)
    newparam = new_image_param(itype=ImageTypes.UNIFORMRANDOM)
    addparam = UniformRandomParam()
    addparam.set_from_datatype(ima1.data.dtype)
    addparam.vmax = int(ima1.data.max() * 0.2)
    panel.new_object(newparam, addparam=addparam, edit=False)

    test_common_operations(panel)

    param = ZCalibrateParam()
    param.a, param.b = 1.2, 0.1
    panel.processor.calibrate(param)

    param = LogP1Param()
    param.n = 1
    panel.processor.compute_logp1(param)

    panel.processor.rotate_90()
    panel.processor.rotate_270()
    panel.processor.flip_horizontally()
    panel.processor.flip_vertically()

    param = RotateParam()
    param.angle = 5.0
    for boundary in param.boundaries[:-1]:
        param.mode = boundary
        panel.processor.rotate_arbitrarily(param)

    param = ResizeParam()
    param.zoom = 1.3
    panel.processor.resize_image(param)

    n = data_size // 10
    panel.processor.extract_roi([[n, n, data_size - n, data_size - n]])

    panel.processor.compute_centroid()
    panel.processor.compute_enclosing_circle()

    data = get_peak2d_data(PeakDataParam(size=data_size))
    ima = create_image("Test image with peaks", data)
    panel.add_object(ima)
    param = PeakDetectionParam()
    param.create_rois = True
    panel.processor.compute_peak_detection(param)

    param = ContourShapeParam()
    panel.processor.compute_contour_shape(param)


def test():
    """Run image unit test scenario 1"""
    with codraft_app_context(save=True) as win:
        test_image_features(win)
        win.imagepanel.open_separate_view((0, 1, 2, 3))


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Datasets
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

import enum
import re
import weakref
from collections import abc
from copy import deepcopy

import guidata.dataset.dataitems as gdi
import guidata.dataset.datatypes as gdt
import numpy as np
from guidata.configtools import get_icon
from guidata.utils import update_dataset
from guiqwt.annotations import AnnotatedCircle
from guiqwt.builder import make
from guiqwt.image import MaskedImageItem
from numpy import ma
from skimage import draw

from codraft.config import Conf, _
from codraft.core.computation.image import scale_data_to_min_max
from codraft.core.model import base


def make_roi_rectangle(x0: int, y0: int, x1: int, y1: int, title: str):
    """Make and return the annnotated rectangle associated to ROI"""
    return make.annotated_rectangle(x0, y0, x1, y1, title)


def make_roi_circle(x0: int, y0: int, x1: int, y1: int, title: str):
    """Make and return the annnotated circle associated to ROI"""
    item = AnnotatedCircle(x0, y0, x1, y1)
    item.annotationparam.title = title
    item.annotationparam.update_annotation(item)
    item.set_style("plot", "shape/drag")
    return item


def to_builtin(obj):
    """Convert an object implementing a numeric value or collection
    into the corresponding builtin/NumPy type.

    Return None if conversion fails."""
    try:
        return int(obj) if int(obj) == float(obj) else float(obj)
    except (TypeError, ValueError):
        pass
    if isinstance(obj, abc.ByteString):
        return str(obj)
    if isinstance(obj, abc.Sequence):
        return str(obj) if len(obj) == len(str(obj)) else list(obj)
    if isinstance(obj, abc.Mapping):
        return dict(obj)
    if isinstance(obj, np.ndarray):
        return obj
    return None


class RoiDataGeometries(enum.Enum):
    """ROI data geometry types"""

    RECTANGLE = 0
    CIRCLE = 1


class RoiDataItem:
    """Object representing an image ROI."""

    def __init__(self, data: np.ndarray):
        self._data = data

    @classmethod
    def from_image(cls, obj, geometry: RoiDataGeometries):
        """Construct roi data item from image object: called for making new ROI items"""
        x0, x1 = obj.x0, obj.size[0] + obj.x0
        if geometry is RoiDataGeometries.RECTANGLE:
            y0, y1 = obj.y0, obj.size[1] + obj.y0
        else:
            y0 = y1 = 0.5 * (2 * obj.y0 + obj.size[1])
        coords = x0, y0, x1, y1
        return cls(coords)

    @property
    def geometry(self) -> RoiDataGeometries:
        """ROI geometry"""
        _x0, y0, _x1, y1 = self._data
        if y0 == y1:
            return RoiDataGeometries.CIRCLE
        return RoiDataGeometries.RECTANGLE

    def get_rect(self):
        """Get rectangle coordinates"""
        x0, y0, x1, y1 = self._data
        if self.geometry is RoiDataGeometries.CIRCLE:
            y0 -= x1 - x0
            y1 += x1 - x0
        return x0, y0, x1, y1

    def get_masked_view(self, data: np.ndarray, maskdata: np.ndarray) -> np.ndarray:
        """Return masked view for data"""
        x0, y0, x1, y1 = self.get_rect()
        masked_view = data.view(ma.MaskedArray)
        masked_view.mask = maskdata
        return masked_view[y0:y1, x0:x1]

    def apply_mask(self, data: np.ndarray, yxratio: float) -> np.ndarray:
        """Apply ROI to data as a mask and return masked array"""
        roi_mask = np.ones_like(data, dtype=bool)
        x0, y0, x1, y1 = self.get_rect()
        if self.geometry is RoiDataGeometries.RECTANGLE:
            roi_mask[y0:y1, x0:x1] = False
        else:
            xc, yc = 0.5 * (x0 + x1), 0.5 * (y0 + y1)
            radius = 0.5 * (x1 - x0)
            rr, cc = draw.ellipse(yc, xc, radius / yxratio, radius, shape=data.shape)
            roi_mask[rr, cc] = False
        return roi_mask

    def make_roi_item(self, index: int, fmt: str, lbl: bool, editable: bool = True):
        """Make ROI plot item"""
        coords = self._data
        if self.geometry is RoiDataGeometries.RECTANGLE:
            func = make_roi_rectangle
        else:
            func = make_roi_circle
        title = "ROI" if index is None else f"ROI{index:02d}"
        return base.make_roi_item(func, coords, title, fmt, lbl, editable)


class ImageParam(gdt.DataSet, base.ObjectItf):
    """Image dataset"""

    CONF_FMT = Conf.view.ima_format
    DEFAULT_FMT = ".1f"
    VALID_DTYPES = (
        np.uint8,
        np.uint16,
        np.int16,
        np.int32,
        np.float32,
        np.float64,
        np.complex128,
    )

    def __init__(self, title=None, comment=None, icon=""):
        gdt.DataSet.__init__(self, title, comment, icon)
        self._dicom_template = None
        self._maskdata_cache = None
        self._roidata_cache = None  # weak reference

    @property
    def size(self):
        """Returns (width, height)"""
        return self.data.shape[1], self.data.shape[0]

    def __add_metadata(self, key, value):
        """Add value to metadata if value can be converted into builtin/NumPy type"""
        stored_val = to_builtin(value)
        if stored_val is not None:
            self.metadata[key] = stored_val

    def set_metadata_from(self, obj):
        """Set metadata from object: dict-like (only string keys are considered)
        or any other object (iterating over supported attributes)"""
        self.metadata = {}
        ptn = r"__[\S_]*__$"
        if isinstance(obj, abc.Mapping):
            for key, value in obj.items():
                if isinstance(key, str) and not re.match(ptn, key):
                    self.__add_metadata(key, value)
        else:
            for attrname in dir(obj):
                if attrname != "GroupLength" and not re.match(ptn, attrname):
                    try:
                        attr = getattr(obj, attrname)
                        if not callable(attr) and attr:
                            self.__add_metadata(attrname, attr)
                    except AttributeError:
                        pass

    @property
    def dicom_template(self):
        """Get DICOM template"""
        return self._dicom_template

    @dicom_template.setter
    def dicom_template(self, template):
        """Set DICOM template"""
        if template is not None:
            ipp = getattr(template, "ImagePositionPatient", None)
            if ipp is not None:
                self.x0, self.y0 = float(ipp[0]), float(ipp[1])
            pxs = getattr(template, "PixelSpacing", None)
            if pxs is not None:
                self.dy, self.dx = float(pxs[0]), float(pxs[1])
            self.set_metadata_from(template)
            self._dicom_template = template

    _tabs = gdt.BeginTabGroup("all")

    _datag = gdt.BeginGroup(_("Data and metadata"))
    data = gdi.FloatArrayItem(_("Data"))
    metadata = base.MetadataItem(_("Metadata"), default={})
    _e_datag = gdt.EndGroup(_("Data and metadata"))

    _dxdyg = gdt.BeginGroup(_("Origin and pixel spacing"))
    _origin = gdt.BeginGroup(_("Origin"))
    x0 = gdi.FloatItem("X<sub>0</sub>", default=0.0)
    y0 = gdi.FloatItem("Y<sub>0</sub>", default=0.0).set_pos(col=1)
    _e_origin = gdt.EndGroup(_("Origin"))
    _pixel_spacing = gdt.BeginGroup(_("Pixel spacing"))
    dx = gdi.FloatItem("Δx", default=1.0, nonzero=True)
    dy = gdi.FloatItem("Δy", default=1.0, nonzero=True).set_pos(col=1)
    _e_pixel_spacing = gdt.EndGroup(_("Pixel spacing"))
    _e_dxdyg = gdt.EndGroup(_("Origin and pixel spacing"))

    _unitsg = gdt.BeginGroup(_("Titles and units"))
    title = gdi.StringItem(_("Image title"), default=_("Untitled"))
    _tabs_u = gdt.BeginTabGroup("units")
    _unitsx = gdt.BeginGroup(_("X-axis"))
    xlabel = gdi.StringItem(_("Title"), default="")
    xunit = gdi.StringItem(_("Unit"), default="")
    _e_unitsx = gdt.EndGroup(_("X-axis"))
    _unitsy = gdt.BeginGroup(_("Y-axis"))
    ylabel = gdi.StringItem(_("Title"), default="")
    yunit = gdi.StringItem(_("Unit"), default="")
    _e_unitsy = gdt.EndGroup(_("Y-axis"))
    _unitsz = gdt.BeginGroup(_("Z-axis"))
    zlabel = gdi.StringItem(_("Title"), default="")
    zunit = gdi.StringItem(_("Unit"), default="")
    _e_unitsz = gdt.EndGroup(_("Z-axis"))
    _e_tabs_u = gdt.EndTabGroup("units")
    _e_unitsg = gdt.EndGroup(_("Titles and units"))

    _e_tabs = gdt.EndTabGroup("all")

    def get_data(self, roi_index: int = None) -> np.ndarray:
        """
        Return original data (if ROI is not defined or `roi_index` is None),
        or ROI data (if both ROI and `roi_index` are defined).

        Returns a masked array.
        """
        if self.roi is None or roi_index is None:
            return self.data
        roidataitem = RoiDataItem(self.roi[roi_index])
        return roidataitem.get_masked_view(self.data, self.maskdata)

    def copy_data_from(self, other, dtype=None):
        """Copy data from other dataset instance"""
        self.x0 = other.x0
        self.y0 = other.y0
        self.dx = other.dx
        self.dy = other.dy
        self.metadata = deepcopy(other.metadata)
        self.data = np.array(other.data, copy=True, dtype=dtype)
        self.dicom_template = other.dicom_template

    def set_data_type(self, dtype):
        """Change data type"""
        self.data = np.array(self.data, dtype=dtype)

    def __viewable_data(self):
        """Return viewable data"""
        data = self.data.real
        if np.any(np.isnan(data)):
            data = np.nan_to_num(data, posinf=0, neginf=0)
        return data

    def __update_item_params(self, data: np.ndarray, item: MaskedImageItem):
        """Update plot item parameters"""
        for axis in ("x", "y", "z"):
            unit = getattr(self, axis + "unit")
            fmt = r"%.1f"
            if unit:
                fmt = r"%.1f (" + unit + ")"
            setattr(item.imageparam, axis + "format", fmt)
        # Updating origin and pixel spacing
        has_origin = self.x0 is not None and self.y0 is not None
        has_pixelspacing = self.dx is not None and self.dy is not None
        if has_origin or has_pixelspacing:
            x0, y0, dx, dy = 0.0, 0.0, 1.0, 1.0
            if has_origin:
                x0, y0 = self.x0, self.y0
            if has_pixelspacing:
                dx, dy = self.dx, self.dy
            item.imageparam.xmin, item.imageparam.xmax = x0, x0 + dx * data.shape[1]
            item.imageparam.ymin, item.imageparam.ymax = y0, y0 + dy * data.shape[0]
        update_dataset(item.imageparam, self.metadata)
        item.imageparam.update_image(item)

    def make_item(self, update_from: MaskedImageItem = None):
        """Make plot item from data"""
        data = self.__viewable_data()
        item = make.maskedimage(
            data,
            self.maskdata,
            title=self.title,
            colormap="jet",
            eliminate_outliers=Conf.view.ima_eliminate_outliers.get(0.1),
            interpolation="nearest",
            show_mask=True,
        )
        if update_from is None:
            self.__update_item_params(data, item)
        else:
            update_dataset(item.imageparam, update_from.imageparam)
            item.imageparam.update_image(item)
        return item

    def update_item(self, item: MaskedImageItem, ref_item: MaskedImageItem = None):
        """Update plot item from data"""
        data = self.__viewable_data()
        item.set_data(data, lut_range=[item.min, item.max])
        item.set_mask(self.maskdata)
        item.imageparam.label = self.title
        if ref_item is not None and Conf.view.ima_ref_lut_range.get(True):
            item.set_lut_range(ref_item.get_lut_range())
        self.__update_item_params(data, item)
        item.plot().update_colormap_axis(item)

    def get_roi_param(self, title, *defaults):
        """Return ROI parameters dataset"""
        roidataitem = RoiDataItem(defaults)

        xd0, yd0, xd1, yd1 = defaults

        def s(name: str, index: int):
            """Returns name<sub>index</sub>"""
            return f"{name}<sub>{index}</sub>"

        if roidataitem.geometry is RoiDataGeometries.RECTANGLE:
            gtitle1 = _("Top left corner")
            gtitle2 = _("Bottom right corner")

            class ROIParam(gdt.DataSet):
                """ROI parameters"""

                geometry = roidataitem.geometry

                def get_suffix(self):
                    """Get suffix text representation for ROI extraction"""
                    return f"x={self.x0}:{self.x1},y={self.y0}:{self.y1}"

                def get_coords(self):
                    """Get ROI coordinates"""
                    return self.x0, self.y0, self.x1, self.y1

                _tlcorner = gdt.BeginGroup(gtitle1)
                x0 = gdi.IntItem(s("X", 0), default=xd0, unit="pixel")
                y0 = gdi.IntItem(s("Y", 0), default=yd0, unit="pixel").set_pos(1)
                _e_tlcorner = gdt.EndGroup(gtitle1)
                _brcorner = gdt.BeginGroup(gtitle2)
                x1 = gdi.IntItem(s("X", 1), default=xd1, unit="pixel")
                y1 = gdi.IntItem(s("Y", 1), default=yd1, unit="pixel").set_pos(1)
                _e_brcorner = gdt.EndGroup(gtitle2)

        else:
            gtitle1 = _("Center coordinates")

            class ROIParam(gdt.DataSet):
                """ROI parameters"""

                geometry = roidataitem.geometry

                def get_single_roi(self):
                    """Get single circular ROI, i.e. after extracting ROI from image"""
                    return np.array([(0, self.r, self.x1 - self.x0, self.r)], int)

                def get_suffix(self):
                    """Get suffix text representation for ROI extraction"""
                    return f"xc={self.xc},yc={self.yc},r={self.r}"

                def get_coords(self):
                    """Get ROI coordinates"""
                    return self.x0, self.yc, self.x1, self.yc

                @property
                def x0(self):
                    """Return rectangle top left corner X coordinate"""
                    return self.xc - self.r

                @property
                def x1(self):
                    """Return rectangle bottom right corner X coordinate"""
                    return self.xc + self.r

                @property
                def y0(self):
                    """Return rectangle top left corner Y coordinate"""
                    return self.yc - self.r

                @property
                def y1(self):
                    """Return rectangle bottom right corner Y coordinate"""
                    return self.yc + self.r

                _tlcorner = gdt.BeginGroup(gtitle1)
                xc = gdi.IntItem(
                    s("X", "C"), default=int(0.5 * (xd0 + xd1)), unit="pixel"
                )
                yc = gdi.IntItem(s("Y", "C"), default=yd0, unit="pixel").set_pos(1)
                _e_tlcorner = gdt.EndGroup(gtitle1)
                r = gdi.IntItem(
                    _("Radius"), default=int(0.5 * (xd1 - xd0)), unit="pixel"
                )

        return ROIParam(title)

    @staticmethod
    def params_to_roidata(params: gdt.DataSetGroup) -> np.ndarray:
        """Convert list of dataset parameters to ROI data"""
        roilist = []
        for roiparam in params.datasets:
            roilist.append(roiparam.get_coords())
        if len(roilist) == 0:
            return None
        return np.array(roilist, int)

    def new_roi_item(self, fmt, lbl, editable, geometry: RoiDataGeometries):
        """Return a new ROI item from scratch"""
        roidataitem = RoiDataItem.from_image(self, geometry)
        return roidataitem.make_roi_item(None, fmt, lbl, editable)

    def roi_coords_to_indexes(self, coords: list) -> np.ndarray:
        """Convert ROI coordinates to indexes"""
        indexes = np.array(coords)
        if indexes.size > 0:
            indexes[:, ::2] -= self.x0 + 0.5 * self.dx
            indexes[:, ::2] /= self.dx
            indexes[:, 1::2] -= self.y0 + 0.5 * self.dy
            indexes[:, 1::2] /= self.dy
        return np.array(indexes, int)

    def iterate_roi_items(self, fmt: str, lbl: bool, editable: bool = True):
        """Iterate over plot items representing Regions of Interest"""
        if self.roi is not None:
            roicoords = np.array(self.roi, float)
            roicoords[:, ::2] *= self.dx
            roicoords[:, ::2] += self.x0 - 0.5 * self.dx
            roicoords[:, 1::2] *= self.dy
            roicoords[:, 1::2] += self.y0 - 0.5 * self.dy
            for index, coords in enumerate(roicoords):
                roidataitem = RoiDataItem(coords)
                yield roidataitem.make_roi_item(index, fmt, lbl, editable)

    @property
    def maskdata(self):
        """Return masked data (areas outside defined regions of interest)"""
        roi_changed = self._roidata_cache is not None and self._roidata_cache() is None
        if self.roi is None:
            if roi_changed:
                self._roidata_cache = None
                self._maskdata_cache = None
        elif roi_changed or self._maskdata_cache is None:
            mask = np.ones_like(self.data, dtype=bool)
            for roirow in self.roi:
                roidataitem = RoiDataItem(roirow)
                roi_mask = roidataitem.apply_mask(self.data, yxratio=self.dy / self.dx)
                mask &= roi_mask
            self._maskdata_cache = mask
            self._roidata_cache = weakref.ref(self.roi)
        return self._maskdata_cache

    def invalidate_maskdata_cache(self):
        """Invalidate mask data cache: force to rebuild it"""
        self._maskdata_cache = None


def create_image(
    title,
    data: np.ndarray = None,
    metadata: dict = None,
    units: tuple = None,
    labels: tuple = None,
):
    """Create a new Image object

    :param str title: image title
    :param numpy.ndarray data: image data
    :param dict metadata: image metadata
    :param tuple units: X, Y, Z units (tuple of strings)
    :param tuple labels: X, Y, Z labels (tuple of strings)
    """
    assert isinstance(title, str)
    assert data is None or isinstance(data, np.ndarray)
    image = ImageParam()
    image.title = title
    image.data = data
    if units is not None:
        image.xunit, image.yunit, image.zunit = units
    if labels is not None:
        image.xlabel, image.ylabel, image.zlabel = labels
    image.metadata = {} if metadata is None else metadata
    # Default visualization settings
    for name, opt in (
        ("colormap", Conf.view.ima_def_colormap),
        ("interpolation", Conf.view.ima_def_interpolation),
    ):
        defval = opt.get(None)
        if defval is not None:
            image.metadata[name] = defval
    # TODO: [P2] Add default signal/image visualization settings
    # 1. Add signal visualization settings?
    # 2. Add more image visualization settings?
    # 3. Add a dialog box to edit default settings in main window
    #    (use a guidata dataset with only a selection of items from guiqwt.styles
    #     classes)
    # 4. Update all active objects when settings were changed
    return image


class ImageDatatypes(base.Choices):
    """Image data types"""

    @classmethod
    def from_dtype(cls, dtype):
        """Return member from NumPy dtype"""
        return getattr(cls, str(dtype).upper(), cls.UINT8)

    @classmethod
    def check(cls):
        """Check if data types are valid"""
        for member in cls:
            assert hasattr(np, member.value)

    UINT8 = enum.auto()
    UINT16 = enum.auto()
    INT16 = enum.auto()
    FLOAT32 = enum.auto()
    FLOAT64 = enum.auto()


ImageDatatypes.check()


class ImageTypes(base.Choices):
    """Image types"""

    ZEROS = _("zeros")
    EMPTY = _("empty")
    GAUSS = _("gaussian")
    UNIFORMRANDOM = _("random (uniform law)")
    NORMALRANDOM = _("random (normal law)")


class ImageParamNew(gdt.DataSet):
    """New image dataset"""

    title = gdi.StringItem(_("Title"), default=_("Untitled"))
    height = gdi.IntItem(
        _("Height"), help=_("Image height (total number of rows)"), min=1, default=500
    )
    width = gdi.IntItem(
        _("Width"), help=_("Image width (total number of columns)"), min=1, default=500
    )
    dtype = gdi.ChoiceItem(
        _("Data type"), ImageDatatypes.get_choices(), default=ImageDatatypes.UINT16
    )
    type = gdi.ChoiceItem(_("Type"), ImageTypes.get_choices())


def new_image_param(title=None, itype=None, height=None, width=None, dtype=None):
    """Create a new Image dataset instance.

    :param str title: dataset title (default: None, uses default title)"""
    if title is None:
        title = _("Untitled image")
    param = ImageParamNew(title=title, icon=get_icon("new_image.svg"))
    param.title = title
    if height is not None:
        param.height = height
    if width is not None:
        param.width = width
    if dtype is not None:
        param.dtype = dtype
    if itype is not None:
        param.type = itype
    return param


IMG_NB = 0


class Gauss2DParam(gdt.DataSet):
    """2D Gaussian parameters"""

    a = gdi.FloatItem("Norm")
    xmin = gdi.FloatItem("Xmin", default=-10).set_pos(col=1)
    sigma = gdi.FloatItem("σ", default=1.0)
    xmax = gdi.FloatItem("Xmax", default=10).set_pos(col=1)
    mu = gdi.FloatItem("μ", default=0.0)
    ymin = gdi.FloatItem("Ymin", default=-10).set_pos(col=1)
    x0 = gdi.FloatItem("X0", default=0)
    ymax = gdi.FloatItem("Ymax", default=10).set_pos(col=1)
    y0 = gdi.FloatItem("Y0", default=0).set_pos(col=0, colspan=1)


def create_image_from_param(newparam, addparam=None, edit=False, parent=None):
    """Create a new Image object from dialog box.

    :param ImageParamNew param: new image parameters
    :param guidata.dataset.datatypes.DataSet addparam: additional parameters
    :param bool edit: Open a dialog box to edit parameters (default: False)
    :param QWidget parent: parent widget
    """
    global IMG_NB  # pylint: disable=global-statement
    if newparam is None:
        newparam = new_image_param()
    incr_sig_nb = not newparam.title
    if incr_sig_nb:
        newparam.title = f"{newparam.title} {IMG_NB + 1:d}"
    if not edit or addparam is not None or newparam.edit(parent=parent):
        if incr_sig_nb:
            IMG_NB += 1
        image = create_image(newparam.title)
        shape = (newparam.height, newparam.width)
        dtype = newparam.dtype.value
        p = addparam
        if newparam.type == ImageTypes.ZEROS:
            image.data = np.zeros(shape, dtype=dtype)
        elif newparam.type == ImageTypes.EMPTY:
            image.data = np.empty(shape, dtype=dtype)
        elif newparam.type == ImageTypes.GAUSS:
            if p is None:
                p = Gauss2DParam(_("New 2D-gaussian image"))
            if p.a is None:
                try:
                    p.a = np.iinfo(dtype).max / 2.0
                except ValueError:
                    p.a = 10.0
            if edit and not p.edit(parent=parent):
                return None
            x, y = np.meshgrid(
                np.linspace(p.xmin, p.xmax, shape[1]),
                np.linspace(p.ymin, p.ymax, shape[0]),
            )
            zgauss = p.a * np.exp(
                -((np.sqrt((x - p.x0) ** 2 + (y - p.y0) ** 2) - p.mu) ** 2)
                / (2.0 * p.sigma**2)
            )
            image.data = np.array(zgauss, dtype=dtype)
        elif newparam.type in (ImageTypes.UNIFORMRANDOM, ImageTypes.NORMALRANDOM):
            pclass = {
                ImageTypes.UNIFORMRANDOM: base.UniformRandomParam,
                ImageTypes.NORMALRANDOM: base.NormalRandomParam,
            }[newparam.type]
            if p is None:
                p = pclass(_("Image") + " - " + newparam.type.value)
                p.set_from_datatype(dtype)
            if edit and not p.edit(parent=parent):
                return None
            rng = np.random.default_rng(p.seed)
            if newparam.type == ImageTypes.UNIFORMRANDOM:
                data = rng.random(shape)
                image.data = scale_data_to_min_max(data, p.vmin, p.vmax)
            elif newparam.type == ImageTypes.NORMALRANDOM:
                image.data = rng.normal(p.mu, p.sigma, size=shape)
            else:
                raise NotImplementedError(f"New param type: {newparam.type.value}")
        return image
    return None



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Computation / Curve fitting module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc

import numpy as np
import scipy.special as sps


# ----- Fitting models ---------------------------------------------------------
class FitModel(abc.ABC):
    """Curve fitting model base class"""

    @classmethod
    @abc.abstractmethod
    def func(cls, x, amp, sigma, x0, y0):
        """Return fitting function"""

    @classmethod
    def get_amp_from_amplitude(
        cls, amplitude, sigma
    ):  # pylint: disable=unused-argument
        """Return amp from function amplitude and sigma"""
        return amplitude

    @classmethod
    def amplitude(cls, amp, sigma):
        """Return function amplitude"""
        return cls.func(0, amp, sigma, 0, 0)

    @classmethod
    @abc.abstractmethod
    def fwhm(cls, amp, sigma):
        """Return function FWHM"""

    @classmethod
    def half_max_segment(cls, amp, sigma, x0, y0):
        """Return segment coordinates for y=half-maximum intersection"""
        hwhm = 0.5 * cls.fwhm(amp, sigma)
        yhm = 0.5 * cls.amplitude(amp, sigma) + y0
        return x0 - hwhm, yhm, x0 + hwhm, yhm


class GaussianModel(FitModel):
    """1-dimensional Gaussian fit model"""

    @classmethod
    def func(cls, x, amp, sigma, x0, y0):
        """Return fitting function"""
        return (
            amp / (sigma * np.sqrt(2 * np.pi)) * np.exp(-0.5 * ((x - x0) / sigma) ** 2)
            + y0
        )

    @classmethod
    def get_amp_from_amplitude(cls, amplitude, sigma):
        """Return amp from function amplitude and sigma"""
        return amplitude * (sigma * np.sqrt(2 * np.pi))

    @classmethod
    def amplitude(cls, amp, sigma):
        """Return function amplitude"""
        return amp / (sigma * np.sqrt(2 * np.pi))

    @classmethod
    def fwhm(cls, amp, sigma):
        """Return function FWHM"""
        return 2 * sigma * np.sqrt(2 * np.log(2))


class LorentzianModel(FitModel):
    """1-dimensional Lorentzian fit model"""

    @classmethod
    def func(cls, x, amp, sigma, x0, y0):
        """Return fitting function"""
        return (amp / (sigma * np.pi)) / (1 + ((x - x0) / sigma) ** 2) + y0

    @classmethod
    def get_amp_from_amplitude(cls, amplitude, sigma):
        """Return amp from function amplitude and sigma"""
        return amplitude * (sigma * np.pi)

    @classmethod
    def amplitude(cls, amp, sigma):
        """Return function amplitude"""
        return amp / (sigma * np.pi)

    @classmethod
    def fwhm(cls, amp, sigma):
        """Return function FWHM"""
        return 2 * sigma


class VoigtModel(FitModel):
    """1-dimensional Voigt fit model"""

    @classmethod
    def func(cls, x, amp, sigma, x0, y0):
        """Return fitting function"""
        # pylint: disable=no-member
        z = (x - x0 + 1j * sigma) / (sigma * np.sqrt(2.0))
        return y0 + amp * sps.wofz(z).real / (sigma * np.sqrt(2 * np.pi))

    @classmethod
    def fwhm(cls, amp, sigma):
        """Return function FWHM"""
        wg = GaussianModel.fwhm(amp, sigma)
        wl = LorentzianModel.fwhm(amp, sigma)
        return 0.5346 * wl + np.sqrt(0.2166 * wl**2 + wg**2)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT MOS07636 HDF5 format support
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
from guidata.utils import update_dataset
from h5py import Group

from codraft.core.io.h5 import common, utils
from codraft.core.model.base import ANN_KEY
from codraft.core.model.image import create_image
from codraft.core.model.signal import create_signal
from codraft.utils.misc import to_string

# Add ignored dataset names
common.NODE_FACTORY.add_ignored_datasets(("PALETTE",))


class BaseMOS07636Node(common.BaseNode):
    """Object representing a HDF5 node, according to MOS07636"""

    ATTR_PATTERN = (None, None)

    def __init__(self, h5file, dset):
        super().__init__(h5file, dset)
        self.xunit = None
        self.yunit = None
        self.zunit = None
        self.xlabel = None
        self.ylabel = None
        self.zlabel = None
        self.__obj_templates = []
        self.__metadata_entries = {}

    def add_object_default_values(self, **template):
        """Add object default values (object template)"""
        self.__obj_templates.append(template)

    def update_from_object_default_values(self, obj):
        """Update object (signal/image) from default values (template), if available"""
        for template in self.__obj_templates:
            update_dataset(obj, template)

    def add_metadata_entry(self, key, value):
        """Add metadata entry to object"""
        self.__metadata_entries[key] = value

    @classmethod
    def match(cls, dset):
        """Return True if h5 dataset match node pattern"""
        name, value = cls.ATTR_PATTERN
        return dset.attrs.get(name) == value

    @property
    def data(self):
        """Data associated to node, if available"""
        if isinstance(self.dset, Group):
            return self.dset["valeur"][()]
        #  This is not a valid dataset according to MOS07636!
        return self.dset[()]

    @property
    def shape_str(self):
        """Return string representation of node shape, if any"""
        try:
            shape = self.data.shape
            if shape:
                return " x ".join([str(size) for size in shape])
        except AttributeError:
            pass
        return ""

    @property
    def dtype_str(self):
        """Return string representation of node data type, if any"""
        try:
            dstr = str(self.data.dtype)
        except AttributeError:
            if isinstance(self.data, (str, bytes)):
                return "string"
            return str(type(self.data))
        if dstr.startswith("|S"):
            return "string"
        return dstr

    @property
    def description(self):
        """Return node description"""
        if isinstance(self.dset, Group):
            desc = utils.process_scalar_value(self.dset, "description", utils.fix_ldata)
            if desc is not None:
                return desc
        return super().description

    def create_object(self):
        """Create native object, if supported"""
        if isinstance(self.dset, Group):
            self.xunit, self.yunit, self.zunit = utils.process_label(self.dset, "unite")
            self.xlabel, self.ylabel, self.zlabel = utils.process_label(
                self.dset, "label"
            )
        for label in ("description", "source"):
            if isinstance(self.dset, Group):
                val = utils.process_scalar_value(self.dset, label, utils.fix_ldata)
                if val is not None:
                    self.metadata[label] = val
        self.metadata.update(self.__metadata_entries)


class ScalarNode(BaseMOS07636Node):
    """Object representing a scalar HDF5 node, according to MOS07636"""

    ATTR_PATTERN = ("CLASS", b"ELEMENTAIRE")

    def __init__(self, h5file, dset):
        super().__init__(h5file, dset)
        if isinstance(self.dset, Group):
            self.xunit = utils.process_scalar_value(self.dset, "unite", self._fix_unit)

    @staticmethod
    def _fix_unit(scdata):
        """Fix unit data"""
        data = scdata[0]
        if not isinstance(data, bytes):
            data = data[0]  # Should not be necessary (invalid format)
        if data == b"NULL":
            return ""
        return utils.fix_ldata(data)

    @property
    def data(self):
        """Data associated to node, if available"""
        try:
            return super().data
        except ValueError:
            #  Handles invalid scalar datasets...
            return self.dset[()]

    @property
    def icon_name(self):
        """Icon name associated to node"""
        return "h5scalar.svg"

    @property
    def text(self):
        """Return node textual representation"""
        text = to_string(self.data)
        suffix = "" if self.xunit is None else " " + self.xunit
        if not text.endswith(suffix):  # Should not be necessary (invalid format)
            text += suffix
        return text


common.NODE_FACTORY.register(ScalarNode)


class SignalNode(BaseMOS07636Node):
    """Object representing a Signal HDF5 node, according to MOS07636"""

    IS_ARRAY = True
    ATTR_PATTERN = ("CLASS", b"COURBE")

    @property
    def icon_name(self):
        """Icon name associated to node"""
        return "signal.svg"

    @property
    def text(self):
        """Return node textual representation"""

    def create_object(self):
        """Create native object, if supported"""
        super().create_object()
        obj = create_signal(
            self.object_title,
            units=(self.xunit, self.yunit),
            labels=(self.xlabel, self.ylabel),
        )
        self.set_signal_data(obj)
        self.update_from_object_default_values(obj)
        return obj


common.NODE_FACTORY.register(SignalNode)


class ImageNode(BaseMOS07636Node):
    """Object representing an Image HDF5 node, according to MOS07636"""

    IS_ARRAY = True
    ATTR_PATTERN = ("CLASS", b"IMAGE")

    @property
    def icon_name(self):
        """Icon name associated to node"""
        return "image.svg"

    @property
    def text(self):
        """Return node textual representation"""

    def create_object(self):
        """Create native object, if supported"""
        super().create_object()
        obj = create_image(
            self.object_title,
            units=(self.xunit, self.yunit, self.zunit),
            labels=(self.xlabel, self.ylabel, self.zlabel),
        )
        self.set_image_data(obj)
        x0, y0 = utils.process_xy_values(self.dset, "origine")
        if x0 is not None and y0 is not None:
            obj.x0, obj.y0 = x0, y0
        dx, dy = utils.process_xy_values(self.dset, "resolution")
        if dx is not None and dy is not None:
            obj.dx, obj.dy = dx, dy
        self.update_from_object_default_values(obj)
        return obj


common.NODE_FACTORY.register(ImageNode)


def handle_margins(node: ImageNode, importer: common.H5Importer):
    """Post-collection trigger handling image margins when available (Vimba Cameras)"""
    try:
        # Vimba Camera HDF5 / node.id: "/Acquisition/AcquisitionBrute"
        margegauche = importer.get_relative(node, "/Parametres_ACQ/MargeGauche", 2)
        margehaute = importer.get_relative(node, "/Parametres_ACQ/MargeHaute", 2)
        binningx = importer.get_relative(node, "/Parametres_ACQ/BinningX", 2)
        binningy = importer.get_relative(node, "/Parametres_ACQ/BinningY", 2)
    except KeyError:
        try:
            # IStar Camera HDF5 / node.id: "/Entrees/Acquisition/AcquisitionBrute"
            margegauche = importer.get_relative(node, "/Parametres_IMG/MargeGauche", 2)
            margehaute = importer.get_relative(node, "/Parametres_IMG/MargeHaute", 2)
            binningx = importer.get_relative(node, "/Parametres_IMG/BinningX", 2)
            binningy = importer.get_relative(node, "/Parametres_IMG/BinningY", 2)
        except KeyError:
            return
    node.add_object_default_values(
        x0=margegauche.data, y0=margehaute.data, dx=binningx.data, dy=binningy.data
    )


common.NODE_FACTORY.add_post_trigger(ImageNode, handle_margins)


def handle_streakcameratimeaxis(node: ImageNode, importer: common.H5Importer):
    """Post-collection trigger handling streak X-axis time conv. when available"""
    try:
        # Streak Camera HDF5 / node.id: "/Acquisition/AcquisitionCorrigee"
        tempspixel = importer.get_relative(node, "/TempsPixel", 1)
        offsettemporel = importer.get_relative(node, "/OffsetTemporel", 1)
    except KeyError:
        return
    if node.id.endswith("AcquisitionCorrigee"):
        node.add_object_default_values(
            x0=offsettemporel.data, dx=tempspixel.data, xunit=tempspixel.xunit
        )


common.NODE_FACTORY.add_post_trigger(ImageNode, handle_streakcameratimeaxis)


def handle_annotations(node: ImageNode, importer: common.H5Importer):
    """Post-collection trigger handling annotations when available"""
    try:
        annotations = importer.get_relative(node, "/Annotations", 1)
    except KeyError:
        return
    node.add_metadata_entry(ANN_KEY, to_string(annotations.data))


common.NODE_FACTORY.add_post_trigger(ImageNode, handle_annotations)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Create dependencies hash
"""

from guidata.configtools import get_module_data_path

from codraft.utils import dephash

dephash.create_dependencies_file(
    get_module_data_path("codraft", "data"), ("guidata", "guiqwt")
)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Memory status widget application test
"""

import numpy as np
import psutil

from codraft.config import Conf
from codraft.core.model.image import Gauss2DParam, ImageTypes, new_image_param
from codraft.tests import codraft_app_context

SHOW = True  # Show test in GUI-based test launcher


def test_memory_alarm(threshold):
    """Memory alarm test"""
    Conf.main.available_memory_threshold.set(threshold)
    rng = np.random.default_rng()
    with codraft_app_context() as win:
        panel = win.imagepanel
        newparam = new_image_param(itype=ImageTypes.GAUSS)
        addparam = Gauss2DParam()
        addparam.x0 = addparam.y0 = rng.integers(-9, 9)
        addparam.sigma = rng.integers(1, 20)
        panel.new_object(newparam, addparam=addparam, edit=False)


def test():
    """Memory alarm test"""
    test_memory_alarm(psutil.virtual_memory().available // (1024**2) * 2)
    test_memory_alarm(psutil.virtual_memory().available // (1024**2) - 100)
    Conf.main.available_memory_threshold.reset()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Base Processor GUI module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import abc
import warnings
from typing import Callable, Dict, List

import guidata.dataset.dataitems as gdi
import guidata.dataset.datatypes as gdt
import numpy as np
from guidata.configtools import get_icon
from guidata.widgets.arrayeditor import ArrayEditor
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from codraft import env
from codraft.config import _
from codraft.core.gui.objectlist import ObjectList
from codraft.core.gui.roieditor import ROIEditorData
from codraft.core.model.base import ResultShape
from codraft.utils import misc
from codraft.utils.qthelpers import create_progress_bar, exec_dialog, qt_try_except


class GaussianParam(gdt.DataSet):
    """Gaussian filter parameters"""

    sigma = gdi.FloatItem("σ", default=1.0)


class MovingAverageParam(gdt.DataSet):
    """Moving average parameters"""

    n = gdi.IntItem(_("Size of the moving window"), default=3, min=1)


class MovingMedianParam(gdt.DataSet):
    """Moving median parameters"""

    n = gdi.IntItem(_("Size of the moving window"), default=3, min=1, even=False)


class ThresholdParam(gdt.DataSet):
    """Threshold parameters"""

    value = gdi.FloatItem(_("Threshold"))


class ClipParam(gdt.DataSet):
    """Data clipping parameters"""

    value = gdi.FloatItem(_("Clipping value"))


class BaseProcessor(QC.QObject):
    """Object handling data processing: operations, processing, computing"""

    SIG_ADD_SHAPE = QC.Signal(int)
    EDIT_ROI_PARAMS = False

    def __init__(self, panel, objlist: ObjectList, plotwidget):
        super().__init__()
        self.panel = panel
        self.objlist = objlist
        self.plotwidget = plotwidget
        self.prefix = panel.PREFIX

    @qt_try_except()
    def compute_sum(self):
        """Compute sum"""
        rows = self.objlist.get_selected_rows()
        outobj = self.panel.create_object()
        outobj.title = "+".join([f"{self.prefix}{row:03d}" for row in rows])
        roilist = []
        for row in rows:
            obj = self.objlist[row]
            if obj.roi is not None:
                roilist.append(obj.roi)
            if outobj.data is None:
                outobj.copy_data_from(obj)
            else:
                outobj.data += np.array(obj.data, dtype=outobj.data.dtype)
                outobj.update_resultshapes_from(obj)
        if roilist:
            outobj.roi = np.vstack(roilist)
        self.panel.add_object(outobj)

    @qt_try_except()
    def compute_average(self):
        """Compute average"""
        rows = self.objlist.get_selected_rows()
        outobj = self.panel.create_object()
        title = ", ".join([f"{self.prefix}{row:03d}" for row in rows])
        outobj.title = f'{_("Average")}({title})'
        original_dtype = self.objlist.get_sel_object().data.dtype
        new_dtype = complex if misc.is_complex_dtype(original_dtype) else float
        roilist = []
        for row in rows:
            obj = self.objlist[row]
            if obj.roi is not None:
                roilist.append(obj.roi)
            if outobj.data is None:
                outobj.copy_data_from(obj, dtype=new_dtype)
            else:
                outobj.data += np.array(obj.data, dtype=outobj.data.dtype)
                outobj.update_resultshapes_from(obj)
        outobj.data /= float(len(rows))
        if misc.is_integer_dtype(original_dtype):
            outobj.set_data_type(dtype=original_dtype)
        if roilist:
            outobj.roi = np.vstack(roilist)
        self.panel.add_object(outobj)

    @qt_try_except()
    def compute_product(self):
        """Compute product"""
        rows = self.objlist.get_selected_rows()
        outobj = self.panel.create_object()
        outobj.title = "*".join([f"{self.prefix}{row:03d}" for row in rows])
        for row in rows:
            obj = self.objlist[row]
            if outobj.data is None:
                outobj.copy_data_from(obj)
            else:
                outobj.data *= np.array(obj.data, dtype=outobj.data.dtype)
        self.panel.add_object(outobj)

    @qt_try_except()
    def compute_difference(self, quad: bool):
        """Compute (quadratic) difference"""
        rows = self.objlist.get_selected_rows()
        outobj = self.panel.create_object()
        outobj.title = "-".join([f"{self.prefix}{row:03d}" for row in rows])
        if quad:
            outobj.title = f"({outobj.title})/sqrt(2)"
        obj0, obj1 = self.objlist.get_sel_object(), self.objlist.get_sel_object(1)
        outobj.copy_data_from(obj0)
        outobj.data -= np.array(obj1.data, dtype=outobj.data.dtype)
        if quad:
            outobj.data = outobj.data / np.sqrt(2.0)
        if np.issubdtype(outobj.data.dtype, np.unsignedinteger):
            outobj.data[obj0.data < obj1.data] = 0
        self.panel.add_object(outobj)

    @qt_try_except()
    def compute_division(self):
        """Compute division"""
        rows = self.objlist.get_selected_rows()
        outobj = self.panel.create_object()
        outobj.title = "/".join([f"{self.prefix}{row:03d}" for row in rows])
        obj0, obj1 = self.objlist.get_sel_object(), self.objlist.get_sel_object(1)
        outobj.copy_data_from(obj0)
        outobj.data = outobj.data / np.array(obj1.data, dtype=outobj.data.dtype)
        self.panel.add_object(outobj)

    def _get_roieditordata(
        self, roidata: np.ndarray = None, singleobj: bool = None
    ) -> ROIEditorData:
        """Eventually open ROI Editing Dialog, and return ROI editor data"""
        # Expected behavior:
        # -----------------
        # * If roidata argument is not None, skip the ROI dialog
        # * If first selected obj has a ROI, use this ROI as default but open
        #   ROI Editor dialog anyway
        # * If multiple objs are selected, then apply the first obj ROI to all

        if roidata is None:
            roieditordata = self.edit_regions_of_interest(
                extract=True, singleobj=singleobj
            )
            if roieditordata is not None and roieditordata.roidata is None:
                # This only happens in unattended mode (forcing QDialog accept)
                return None
        else:
            roieditordata = ROIEditorData(roidata=roidata, singleobj=singleobj)
        return roieditordata

    @abc.abstractmethod
    def extract_roi(self, roidata: np.ndarray = None) -> None:
        """Extract Region Of Interest (ROI) from data"""

    @abc.abstractmethod
    def swap_axes(self):
        """Swap data axes"""

    @abc.abstractmethod
    def compute_abs(self):
        """Compute absolute value"""

    @abc.abstractmethod
    def compute_log10(self):
        """Compute Log10"""

    # ------Data Processing
    @abc.abstractmethod
    def apply_11_func(self, obj, orig, func, param, message):
        """Apply 11 function: 1 object in --> 1 object out"""

    def compute_11(
        self,
        name: str,
        func: Callable,
        param: gdt.DataSet = None,
        suffix: Callable = None,
        func_obj: Callable = None,
        edit: bool = True,
    ):
        """Compute 11 function: 1 object in --> 1 object out"""
        if param is not None:
            if edit and not param.edit(parent=self.panel.parent()):
                return
        self._compute_11_subroutine([name], func, [param], suffix, func_obj)

    def compute_1n(
        self,
        names: List,
        func: Callable,
        params: List = None,
        suffix: Callable = None,
        func_obj: Callable = None,
        edit: bool = True,
    ):
        """Compute 1n function: 1 object in --> n objects out"""
        if params is not None:
            group = gdt.DataSetGroup(params, title=_("Parameters"))
            if edit and not group.edit(parent=self.panel.parent()):
                return
        self._compute_11_subroutine(names, func, params, suffix, func_obj)

    def _compute_11_subroutine(
        self,
        names: List,
        func: Callable,
        params: List,
        suffix: Callable,
        func_obj: Callable,
    ):
        """Compute 11 subroutine: used by compute 11 and compute 1n methods"""
        rows = self.objlist.get_selected_rows()
        with create_progress_bar(
            self.panel, names[0], max_=len(rows) * len(params)
        ) as progress:
            for i_row, row in enumerate(rows):
                for i_param, (param, name) in enumerate(zip(params, names)):
                    progress.setValue(i_row * i_param)
                    progress.setLabelText(name)
                    QW.QApplication.processEvents()
                    if progress.wasCanceled():
                        break
                    orig = self.objlist[row]
                    obj = self.panel.create_object()
                    obj.title = f"{name}({self.prefix}{row:03d})"
                    if suffix is not None:
                        obj.title += "|" + suffix(param)
                    obj.copy_data_from(orig)
                    message = _("Computing:") + " " + obj.title
                    self.apply_11_func(obj, orig, func, param, message)
                    if func_obj is not None:
                        if param is None:
                            func_obj(obj)
                        else:
                            func_obj(obj, param)
                    self.panel.add_object(obj)

    @abc.abstractmethod
    def apply_10_func(self, orig, func, param, message) -> ResultShape:
        """Apply 10 function: 1 object in --> 0 object out (scalar result)"""

    def compute_10(
        self,
        name: str,
        func: Callable,
        param: gdt.DataSet = None,
        suffix: Callable = None,
        edit: bool = True,
    ) -> Dict[int, ResultShape]:
        """Compute 10 function: 1 object in --> 0 object out
        (the result of this method is stored in original object's metadata)"""
        if param is not None:
            if edit and not param.edit(parent=self.panel.parent()):
                return None
        rows = self.objlist.get_selected_rows()
        with create_progress_bar(self.panel, name, max_=len(rows)) as progress:
            results = {}
            xlabels = None
            ylabels = []
            title_suffix = "" if suffix is None else "|" + suffix(param)
            for idx, row in enumerate(rows):
                progress.setValue(idx)
                QW.QApplication.processEvents()
                if progress.wasCanceled():
                    break
                orig = self.objlist[row]
                title = f"{name}{title_suffix}"
                message = _("Computing:") + " " + title
                result = self.apply_10_func(orig, func, param, message)
                if result is None:
                    continue
                results[row] = result
                xlabels = result.xlabels
                self.SIG_ADD_SHAPE.emit(row)
                self.panel.selection_changed()
                self.panel.SIG_UPDATE_PLOT_ITEM.emit(row)
                for _i_row_res in range(result.array.shape[0]):
                    ylabel = f"{name}({self.prefix}{idx:03d}){title_suffix}"
                    ylabels.append(ylabel)
        if results:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                dlg = ArrayEditor(self.panel.parent())
                title = _("Results")
                res = np.vstack([result.array for result in results.values()])
                dlg.setup_and_check(
                    res, title, readonly=True, xlabels=xlabels, ylabels=ylabels
                )
                dlg.setObjectName(f"{self.prefix}_results")
                dlg.resize(750, 300)
                exec_dialog(dlg)
        return results

    @abc.abstractmethod
    @qt_try_except()
    def calibrate(self, param=None) -> None:
        """Compute data linear calibration"""

    @staticmethod
    @abc.abstractmethod
    def func_gaussian_filter(x, y, p):
        """Compute gaussian filter"""

    @qt_try_except()
    def compute_gaussian(self, param: GaussianParam = None) -> None:
        """Compute gaussian filter"""
        edit = param is None
        if edit:
            param = GaussianParam(_("Gaussian filter"))
        func = self.func_gaussian_filter
        self.compute_11(
            "GaussianFilter",
            func,
            param,
            suffix=lambda p: f"σ={p.sigma:.3f} pixels",
            edit=edit,
        )

    @staticmethod
    @abc.abstractmethod
    def func_moving_average(x, y, p):
        """Moving average computing function"""

    @qt_try_except()
    def compute_moving_average(self, param: MovingAverageParam = None) -> None:
        """Compute moving average"""
        edit = param is None
        if edit:
            param = MovingAverageParam(_("Moving average"))
        func = self.func_moving_average
        self.compute_11("MovAvg", func, param, suffix=lambda p: f"n={p.n}", edit=edit)

    @staticmethod
    @abc.abstractmethod
    def func_moving_median(x, y, p):
        """Moving median computing function"""

    @qt_try_except()
    def compute_moving_median(self, param: MovingMedianParam = None) -> None:
        """Compute moving median"""
        edit = param is None
        if edit:
            param = MovingMedianParam(_("Moving median"))
        func = self.func_moving_median
        self.compute_11("MovMed", func, param, suffix=lambda p: f"n={p.n}", edit=edit)

    @abc.abstractmethod
    @qt_try_except()
    def compute_wiener(self):
        """Compute Wiener filter"""

    @abc.abstractmethod
    @qt_try_except()
    def compute_fft(self):
        """Compute iFFT"""

    @abc.abstractmethod
    @qt_try_except()
    def compute_ifft(self):
        """Compute FFT"""

    # ------Computing
    def edit_regions_of_interest(
        self, extract: bool = False, singleobj: bool = None
    ) -> ROIEditorData:
        """Define Region Of Interest (ROI) for computing functions"""
        roieditordata = self.panel.get_roi_dialog(extract=extract, singleobj=singleobj)
        if roieditordata is not None:
            row = self.objlist.get_selected_rows()[0]
            obj = self.objlist[row]
            roigroup = obj.roidata_to_params(roieditordata.roidata)
            if (
                env.execenv.unattended
                or roieditordata.roidata.size == 0
                or not self.EDIT_ROI_PARAMS
                or roigroup.edit(parent=self.panel)
            ):
                roidata = obj.params_to_roidata(roigroup)
                if roieditordata.modified:
                    # If ROI has been modified, save ROI (even in "extract mode")
                    obj.roi = roidata
                    self.SIG_ADD_SHAPE.emit(row)
                    self.panel.selection_changed()
                    self.panel.SIG_UPDATE_PLOT_ITEMS.emit()
        return roieditordata

    def delete_regions_of_interest(self):
        """Delete Regions Of Interest"""
        for row in self.objlist.get_selected_rows():
            obj = self.objlist[row]
            if obj.roi is not None:
                obj.roi = None
                self.panel.selection_changed()
                self.panel.SIG_UPDATE_PLOT_ITEMS.emit()

    @abc.abstractmethod
    def _get_stat_funcs(self):
        """Return statistics functions list"""

    @qt_try_except()
    def compute_stats(self):
        """Compute data statistics"""
        row = self.objlist.get_selected_rows()[0]
        obj = self.objlist.get_sel_object()
        stfuncs = self._get_stat_funcs()
        nbcal = len(stfuncs)
        roi_nb = 0 if obj.roi is None else obj.roi.shape[0]
        res = np.zeros((1 + roi_nb, nbcal))
        xlabels = [None] * nbcal
        obj_t = f"{self.prefix}{row:03d}"
        ylabels = [None] * (roi_nb + 1)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            with np.errstate(all="ignore"):
                for iroi, roi_index in enumerate([None] + list(range(roi_nb))):
                    for ical, (label, func) in enumerate(stfuncs):
                        xlabels[ical] = label
                        res[iroi, ical] = func(obj.get_data(roi_index=roi_index))
                    if roi_index is None:
                        ylabels[iroi] = obj_t
                    else:
                        ylabels[iroi] = f"{obj_t}|ROI{roi_index:02d}"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            dlg = ArrayEditor(self.panel.parent())
            title = _("Statistics")
            dlg.setup_and_check(
                res, title, readonly=True, xlabels=xlabels, ylabels=ylabels
            )
            dlg.setObjectName(f"{self.prefix}_stats")
            dlg.setWindowIcon(get_icon("stats.svg"))
            dlg.resize(750, 300)
            exec_dialog(dlg)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
I/O test

Testing CodraFT specific formats.
"""

from codraft.core.io.image import (
    FXDFile,
    SCORFile,
    SIFFile,
    imread_fxd,
    imread_scor,
    imread_sif,
)
from codraft.core.io.signal import read_signal
from codraft.env import execenv
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.tests import try_open_test_data
from codraft.utils.vistools import view_curve_items, view_images

SHOW = True  # Show test in GUI-based test launcher


@try_open_test_data("Testing CSV file reader", "*.txt")
def test_csv(fname=None):
    """Testing CSV files"""
    signal = read_signal(fname)
    execenv.print(signal)
    view_curve_items([signal.make_item()])


@try_open_test_data("Testing SIF file handler", "*.sif")
def test_sif(fname=None):
    """Testing SIF files"""
    execenv.print(SIFFile(fname))
    data = imread_sif(fname)[0]
    view_images(data)


@try_open_test_data("Testing FXD file handler", "*.fxd")
def test_fxd(fname=None):
    """Testing FXD files"""
    execenv.print(FXDFile(fname))
    data = imread_fxd(fname)
    view_images(data)


@try_open_test_data("Testing SCOR-DATA file handler", "*.scor-data")
def test_scordata(fname=None):
    """Testing SCOR-DATA files"""
    execenv.print(SCORFile(fname))
    data = imread_scor(fname)
    view_images(data)


def io_test():
    """I/O test"""
    with qt_app_context():
        test_csv()
        test_sif()
        test_fxd()
        test_scordata()


if __name__ == "__main__":
    io_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Image Processor GUI module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...


import numpy as np
import scipy.ndimage as spi
import scipy.signal as sps
from guidata.dataset.dataitems import BoolItem, ChoiceItem, FloatItem, IntItem
from guidata.dataset.datatypes import DataSet, DataSetGroup, ValueProp
from guiqwt.widgets.resizedialog import ResizeDialog
from numpy import ma
from qtpy import QtWidgets as QW

from codraft.config import APP_NAME, _
from codraft.core.computation.image import (
    distance_matrix,
    flatfield,
    get_2d_peaks_coords,
    get_centroid_fourier,
    get_contour_shapes,
    get_enclosing_circle,
)
from codraft.core.gui.processor.base import BaseProcessor, ClipParam, ThresholdParam
from codraft.core.model.base import BaseProcParam, ResultShape, ShapeTypes
from codraft.core.model.image import ImageParam, RoiDataGeometries, RoiDataItem
from codraft.utils.qthelpers import create_progress_bar, qt_try_except


class LogP1Param(DataSet):
    """Log10 parameters"""

    n = FloatItem("n")


class RotateParam(DataSet):
    """Rotate parameters"""

    boundaries = ("constant", "nearest", "reflect", "wrap")
    prop = ValueProp(False)

    angle = FloatItem(f"{_('Angle')} (°)")
    mode = ChoiceItem(
        _("Mode"), list(zip(boundaries, boundaries)), default=boundaries[0]
    )
    cval = FloatItem(
        _("cval"),
        default=0.0,
        help=_(
            "Value used for points outside the "
            "boundaries of the input if mode is "
            "'constant'"
        ),
    )
    reshape = BoolItem(
        _("Reshape the output array"),
        default=True,
        help=_(
            "Reshape the output array "
            "so that the input array is "
            "contained completely in the output"
        ),
    )
    prefilter = BoolItem(_("Prefilter the input image"), default=True).set_prop(
        "display", store=prop
    )
    order = IntItem(
        _("Order"),
        default=3,
        min=0,
        max=5,
        help=_("Spline interpolation order"),
    ).set_prop("display", active=prop)


class ResizeParam(DataSet):
    """Resize parameters"""

    boundaries = ("constant", "nearest", "reflect", "wrap")
    prop = ValueProp(False)

    zoom = FloatItem(_("Zoom"))
    mode = ChoiceItem(
        _("Mode"), list(zip(boundaries, boundaries)), default=boundaries[0]
    )
    cval = FloatItem(
        _("cval"),
        default=0.0,
        help=_(
            "Value used for points outside the "
            "boundaries of the input if mode is "
            "'constant'"
        ),
    )
    prefilter = BoolItem(_("Prefilter the input image"), default=True).set_prop(
        "display", store=prop
    )
    order = IntItem(
        _("Order"),
        default=3,
        min=0,
        max=5,
        help=_("Spline interpolation order"),
    ).set_prop("display", active=prop)


class FlatFieldParam(BaseProcParam):
    """Flat-field parameters"""

    threshold = FloatItem(_("Threshold"), default=0.0)


class ZCalibrateParam(DataSet):
    """Image linear calibration parameters"""

    a = FloatItem("a", default=1.0)
    b = FloatItem("b", default=0.0)


class GenericDetectionParam(DataSet):
    """Generic detection parameters"""

    threshold = FloatItem(
        _("Relative threshold"),
        default=0.5,
        min=0.1,
        max=0.9,
        help=_(
            "Detection threshold, relative to difference between "
            "data maximum and minimum"
        ),
    )


class PeakDetectionParam(GenericDetectionParam):
    """Peak detection parameters"""

    size = IntItem(
        _("Neighborhoods size"),
        default=10,
        min=1,
        unit="pixels",
        help=_(
            "Size of the sliding window used in maximum/minimum filtering algorithm"
        ),
    )
    create_rois = BoolItem(_("Create regions of interest"), default=True)


class ContourShapeParam(GenericDetectionParam):
    """Contour shape parameters"""

    shapes = (
        ("ellipse", _("Ellipse")),
        ("circle", _("Circle")),
    )
    shape = ChoiceItem(_("Shape"), shapes, default="ellipse")


class ImageProcessor(BaseProcessor):
    """Object handling image processing: operations, processing, computing"""

    # pylint: disable=duplicate-code

    EDIT_ROI_PARAMS = True

    def compute_logp1(self, param: LogP1Param = None) -> None:
        """Compute base 10 logarithm"""
        edit = param is None
        if edit:
            param = LogP1Param("Log10(z+n)")
        self.compute_11(
            "Log10(z+n)",
            lambda z, p: np.log10(z + p.n),
            param,
            suffix=lambda p: f"n={p.n}",
            edit=edit,
        )

    def rotate_arbitrarily(self, param: RotateParam = None) -> None:
        """Rotate data arbitrarily"""
        edit = param is None
        if edit:
            param = RotateParam(_("Rotation"))
        # TODO: [P2] Instead of removing geometric shapes, apply rotation
        self.compute_11(
            "Rotate",
            lambda x, p: spi.rotate(
                x,
                p.angle,
                reshape=p.reshape,
                order=p.order,
                mode=p.mode,
                cval=p.cval,
                prefilter=p.prefilter,
            ),
            param,
            suffix=lambda p: f"α={p.angle:.3f}°, mode='{p.mode}'",
            func_obj=lambda obj, _param: obj.remove_resultshapes(),
            edit=edit,
        )

    def rotate_90(self):
        """Rotate data 90°"""
        # TODO: [P2] Instead of removing geometric shapes, apply 90° rotation
        self.compute_11(
            "Rotate90",
            np.rot90,
            func_obj=lambda obj: obj.remove_resultshapes(),
        )

    def rotate_270(self):
        """Rotate data 270°"""
        # TODO: [P2] Instead of removing geometric shapes, apply 270° rotation
        self.compute_11(
            "Rotate270",
            lambda x: np.rot90(x, 3),
            func_obj=lambda obj: obj.remove_resultshapes(),
        )

    def flip_horizontally(self):
        """Flip data horizontally"""
        # TODO: [P2] Instead of removing geometric shapes, apply horizontal flip
        self.compute_11(
            "HFlip",
            np.fliplr,
            func_obj=lambda obj: obj.remove_resultshapes(),
        )

    def flip_vertically(self):
        """Flip data vertically"""
        # TODO: [P2] Instead of removing geometric shapes, apply vertical flip
        self.compute_11(
            "VFlip",
            np.flipud,
            func_obj=lambda obj: obj.remove_resultshapes(),
        )

    def resize_image(self, param: ResizeParam = None) -> None:
        """Resize image"""
        obj0 = self.objlist.get_sel_object(0)
        for obj in self.objlist.get_sel_objects():
            if obj.size != obj0.size:
                QW.QMessageBox.warning(
                    self.panel.parent(),
                    APP_NAME,
                    _("Warning:")
                    + "\n"
                    + _("Selected images do not have the same size"),
                )

        edit = param is None
        if edit:
            original_size = obj0.size
            dlg = ResizeDialog(
                self.plotwidget,
                new_size=original_size,
                old_size=original_size,
                text=_("Destination size:"),
            )
            if not dlg.exec():
                return
            param = ResizeParam(_("Resize"))
            param.zoom = dlg.get_zoom()

        def func_obj(obj, param):
            """Zooming function"""
            if obj.dx is not None and obj.dy is not None:
                obj.dx, obj.dy = obj.dx / param.zoom, obj.dy / param.zoom
            # TODO: [P2] Instead of removing geometric shapes, apply zoom
            obj.remove_resultshapes()

        self.compute_11(
            "Zoom",
            lambda x, p: spi.interpolation.zoom(
                x,
                p.zoom,
                order=p.order,
                mode=p.mode,
                cval=p.cval,
                prefilter=p.prefilter,
            ),
            param,
            suffix=lambda p: f"zoom={p.zoom:.3f}",
            func_obj=func_obj,
            edit=edit,
        )

    def extract_roi(self, roidata: np.ndarray = None, singleobj: bool = None) -> None:
        """Extract Region Of Interest (ROI) from data"""
        roieditordata = self._get_roieditordata(roidata, singleobj)
        if roieditordata is None or roieditordata.is_empty:
            return
        obj = self.objlist.get_sel_object()
        group = obj.roidata_to_params(roieditordata.roidata)

        if roieditordata.singleobj:

            def suffix_func(group: DataSetGroup):
                if len(group.datasets) == 1:
                    p = group.datasets[0]
                    return p.get_suffix()
                return ""

            def extract_roi_func(data: np.ndarray, group: DataSetGroup):
                """Extract ROI function on data"""
                if len(group.datasets) == 1:
                    p = group.datasets[0]
                    return data.copy()[p.y0 : p.y1, p.x0 : p.x1]
                out = np.zeros_like(data)
                for p in group.datasets:
                    slice1, slice2 = slice(p.y0, p.y1 + 1), slice(p.x0, p.x1 + 1)
                    out[slice1, slice2] = data[slice1, slice2]
                x0 = min([p.x0 for p in group.datasets])
                y0 = min([p.y0 for p in group.datasets])
                x1 = max([p.x1 for p in group.datasets])
                y1 = max([p.y1 for p in group.datasets])
                return out[y0:y1, x0:x1]

            def extract_roi_func_obj(image: ImageParam, group: DataSetGroup):
                """Extract ROI function on object"""
                image.x0 += min([p.x0 for p in group.datasets])
                image.y0 += min([p.y0 for p in group.datasets])
                image.remove_resultshapes()

            # TODO: [P2] Instead of removing geometric shapes, apply ROI extract
            self.compute_11(
                "ROI",
                extract_roi_func,
                group,
                suffix=suffix_func,
                func_obj=extract_roi_func_obj,
                edit=False,
            )

        else:

            def extract_roi_func_obj(image: ImageParam, p: DataSet):
                """Extract ROI function on object"""
                image.x0 += p.x0
                image.y0 += p.y0
                image.remove_resultshapes()
                if p.geometry is RoiDataGeometries.CIRCLE:
                    # Circular ROI
                    image.roi = p.get_single_roi()

            # TODO: [P2] Instead of removing geometric shapes, apply roi extract
            self.compute_1n(
                [f"ROI{iroi}" for iroi in range(len(group.datasets))],
                lambda z, p: z.copy()[p.y0 : p.y1, p.x0 : p.x1],
                group.datasets,
                suffix=lambda p: p.get_suffix(),
                func_obj=extract_roi_func_obj,
                edit=False,
            )

    def swap_axes(self):
        """Swap data axes"""
        self.compute_11(
            "SwapAxes",
            lambda z: z.T,
            func_obj=lambda obj: obj.remove_resultshapes(),
        )

    def compute_abs(self):
        """Compute absolute value"""
        self.compute_11("Abs", np.abs)

    def compute_log10(self):
        """Compute Log10"""
        self.compute_11("Log10", np.log10)

    @qt_try_except()
    def flat_field_correction(self, param: FlatFieldParam = None) -> None:
        """Compute flat field correction"""
        edit = param is None
        rawdata = self.objlist.get_sel_object().data
        flatdata = self.objlist.get_sel_object(1).data
        if edit:
            param = FlatFieldParam(_("Flat field"))
            param.set_from_datatype(rawdata.dtype)
        if not edit or param.edit(self.panel.parent()):
            rows = self.objlist.get_selected_rows()
            robj = self.panel.create_object()
            robj.title = (
                "FlatField("
                + (",".join([f"{self.prefix}{row:03d}" for row in rows]))
                + f",threshold={param.threshold})"
            )
            robj.data = flatfield(rawdata, flatdata, param.threshold)
            self.panel.add_object(robj)

    # ------Image Processing
    def apply_11_func(self, obj, orig, func, param, message):
        """Apply 11 function: 1 object in --> 1 object out"""

        # (self is used by @qt_try_except)
        # pylint: disable=unused-argument
        @qt_try_except(message)
        def apply_11_func_callback(self, obj, orig, func, param):
            """Apply 11 function callback: 1 object in --> 1 object out"""
            if param is None:
                obj.data = func(orig.data)
            else:
                obj.data = func(orig.data, param)

        return apply_11_func_callback(self, obj, orig, func, param)

    @qt_try_except()
    def calibrate(self, param: ZCalibrateParam = None) -> None:
        """Compute data linear calibration"""
        edit = param is None
        if edit:
            param = ZCalibrateParam(_("Linear calibration"), "y = a.x + b")
        self.compute_11(
            "LinearCal",
            lambda x, p: p.a * x + p.b,
            param,
            suffix=lambda p: "z={p.a}*z+{p.b}",
            edit=edit,
        )

    @qt_try_except()
    def compute_threshold(self, param: ThresholdParam = None) -> None:
        """Compute threshold clipping"""
        edit = param is None
        if edit:
            param = ThresholdParam(_("Thresholding"))
        self.compute_11(
            "Threshold",
            lambda x, p: np.clip(x, p.value, x.max()),
            param,
            suffix=lambda p: f"min={p.value} lsb",
            edit=edit,
        )

    @qt_try_except()
    def compute_clip(self, param: ClipParam = None) -> None:
        """Compute maximum data clipping"""
        edit = param is None
        if edit:
            param = ClipParam(_("Clipping"))
        self.compute_11(
            "Clip",
            lambda x, p: np.clip(x, x.min(), p.value),
            param,
            suffix=lambda p: f"max={p.value} lsb",
            edit=edit,
        )

    @staticmethod
    def func_gaussian_filter(x, p):  # pylint: disable=arguments-differ
        """Compute gaussian filter"""
        return spi.gaussian_filter(x, p.sigma)

    @qt_try_except()
    def compute_fft(self):
        """Compute FFT"""
        self.compute_11("FFT", np.fft.fft2)

    @qt_try_except()
    def compute_ifft(self):
        "Compute iFFT" ""
        self.compute_11("iFFT", np.fft.ifft2)

    @staticmethod
    def func_moving_average(x, p):  # pylint: disable=arguments-differ
        """Moving average computing function"""
        return spi.uniform_filter(x, size=p.n, mode="constant")

    @staticmethod
    def func_moving_median(x, p):  # pylint: disable=arguments-differ
        """Moving median computing function"""
        return sps.medfilt(x, kernel_size=p.n)

    @qt_try_except()
    def compute_wiener(self):
        """Compute Wiener filter"""
        self.compute_11("WienerFilter", sps.wiener)

    # ------Image Computing
    def apply_10_func(self, orig, func, param, message) -> ResultShape:
        """Apply 10 function: 1 object in --> 0 object out (scalar result)"""

        # (self is used by @qt_try_except)
        # pylint: disable=unused-argument
        @qt_try_except(message)
        def apply_10_func_callback(self, orig, func, param):
            """Apply 10 function cb: 1 object in --> 0 object out (scalar result)"""
            if param is None:
                return func(orig)
            return func(orig, param)

        return apply_10_func_callback(self, orig, func, param)

    @staticmethod
    def __apply_origin_size_roi(image, func, *args) -> np.ndarray:
        """Exec computation taking into account image x0, y0, dx, dy and ROIs"""
        res = []
        for i_roi in image.iterate_roi_indexes():
            coords = func(image.get_data(i_roi), *args)
            if coords.size:
                if image.roi is not None:
                    x0, y0, _x1, _y1 = RoiDataItem(image.roi[i_roi]).get_rect()
                    coords[:, ::2] += x0
                    coords[:, 1::2] += y0
                coords[:, ::2] = image.dx * coords[:, ::2] + image.x0
                coords[:, 1::2] = image.dy * coords[:, 1::2] + image.y0
                idx = np.ones((coords.shape[0], 1)) * i_roi
                coords = np.hstack([idx, coords])
                res.append(coords)
        if res:
            return np.vstack(res)
        return None

    @qt_try_except()
    def compute_centroid(self):
        """Compute image centroid"""

        def get_centroid_coords(data: np.ndarray):
            """Return centroid coordinates"""
            y, x = get_centroid_fourier(data)
            return np.array([(x, y)])

        def centroid(image: ImageParam):
            """Compute centroid"""
            res = self.__apply_origin_size_roi(image, get_centroid_coords)
            if res is not None:
                return image.add_resultshape("Centroid", ShapeTypes.MARKER, res)
            return None

        self.compute_10(_("Centroid"), centroid)

    @qt_try_except()
    def compute_enclosing_circle(self):
        """Compute minimum enclosing circle"""

        def get_enclosing_circle_coords(data: np.ndarray):
            """Return diameter coords for the circle contour enclosing image
            values above threshold (FWHM)"""
            x, y, r = get_enclosing_circle(data)
            return np.array([[x - r, y, x + r, y]])

        def enclosing_circle(image: ImageParam):
            """Compute minimum enclosing circle"""
            res = self.__apply_origin_size_roi(image, get_enclosing_circle_coords)
            if res is not None:
                return image.add_resultshape("MinEnclosCircle", ShapeTypes.CIRCLE, res)
            return None

        # TODO: [P2] Find a way to add the circle to the computing results
        #  as in "enclosingcircle_test.py"
        self.compute_10(_("MinEnclosingCircle"), enclosing_circle)

    @qt_try_except()
    def compute_peak_detection(self, param: PeakDetectionParam = None) -> None:
        """Compute 2D peak detection"""

        def peak_detection(image: ImageParam, p: PeakDetectionParam):
            """Compute centroid"""
            res = self.__apply_origin_size_roi(
                image, get_2d_peaks_coords, p.size, p.threshold
            )
            if res is not None:
                return image.add_resultshape("Peaks", ShapeTypes.POINT, res)
            return None

        edit = param is None
        if edit:
            data = self.objlist.get_sel_object().data
            param = PeakDetectionParam()
            param.size = max(min(data.shape) // 40, 50)

        results = self.compute_10(_("Peaks"), peak_detection, param, edit=edit)
        if param.create_rois:
            with create_progress_bar(
                self.panel, _("Create regions of interest"), max_=len(results)
            ) as progress:
                for idx, (row, result) in enumerate(results.items()):
                    progress.setValue(idx)
                    QW.QApplication.processEvents()
                    if progress.wasCanceled():
                        break
                    obj = self.objlist[row]
                    dist = distance_matrix(result.data)
                    dist_min = dist[dist != 0].min()
                    assert dist_min > 0
                    radius = int(0.5 * dist_min / np.sqrt(2) - 1)
                    assert radius >= 1
                    roicoords = []
                    ymax, xmax = obj.data.shape
                    for x, y in result.data:
                        coords = [
                            max(x - radius, 0),
                            max(y - radius, 0),
                            min(x + radius, xmax),
                            min(y + radius, ymax),
                        ]
                        roicoords.append(coords)
                    obj.roi = np.array(roicoords, int)
                    self.SIG_ADD_SHAPE.emit(row)
                    self.panel.selection_changed()
                    self.panel.SIG_UPDATE_PLOT_ITEM.emit(row)

    @qt_try_except()
    def compute_contour_shape(self, param: ContourShapeParam = None) -> None:
        """Compute contour shape fit"""

        def contour_shape(image: ImageParam, p: ContourShapeParam):
            """Compute contour shape fit"""
            res = self.__apply_origin_size_roi(
                image, get_contour_shapes, p.shape, p.threshold
            )
            if res is not None:
                shape = ShapeTypes.CIRCLE if p.shape == "circle" else ShapeTypes.ELLIPSE
                return image.add_resultshape("Contour", shape, res)
            return None

        edit = param is None
        if edit:
            param = ContourShapeParam()
        self.compute_10(_("Contour"), contour_shape, param, edit=edit)

    def _get_stat_funcs(self):
        """Return statistics functions list"""
        # Be careful to use systematically functions adapted to masked arrays
        # (e.g. numpy.ma median, and *not* numpy.median)
        return [
            ("min(z)", lambda z: z.min()),
            ("max(z)", lambda z: z.max()),
            ("<z>", lambda z: z.mean()),
            ("Median(z)", ma.median),
            ("σ(z)", lambda z: z.std()),
            ("Σ(z)", lambda z: z.sum()),
            ("<z>/σ(z)", lambda z: z.mean() / z.std()),
        ]



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
HDF5 browser test 2

Testing for memory leak
"""

import os

import numpy as np
import psutil

from codraft.env import execenv
from codraft.tests.data import get_test_fnames
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.vistools import view_curves
from codraft.widgets.h5browser import H5BrowserDialog

SHOW = True  # Show test in GUI-based test launcher


def memoryleak_test(fname, iterations=20):
    """Memory leak test"""
    with qt_app_context():
        proc = psutil.Process(os.getpid())
        fname = get_test_fnames(fname)[0]
        dlg = H5BrowserDialog(None)
        memlist = []
        for i in range(iterations):
            dlg.setup(fname)
            memdata = proc.memory_info().vms / 1024**2
            memlist.append(memdata)
            execenv.print(i + 1, ":", memdata, "MB")
            dlg.browser.tree.select_all(True)
            dlg.browser.tree.toggle_all(True)
            execenv.print(i + 1, ":", proc.memory_info().vms / 1024**2, "MB")
            dlg.show()
            dlg.accept()
            dlg.close()
            execenv.print(i + 1, ":", proc.memory_info().vms / 1024**2, "MB")
            dlg.cleanup()
        view_curves(
            np.array(memlist),
            title="Memory leak test for HDF5 browser dialog",
            ylabel="Memory (MB)",
        )


if __name__ == "__main__":
    memoryleak_test("scenario*.h5")



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Unit test scenario: all features

Maximizing test coverage.
"""

# pylint: disable=duplicate-code

from codraft.tests import codraft_app_context
from codraft.tests.scenario_ima_app import test_image_features
from codraft.tests.scenario_sig_app import test_signal_features


def test():
    """Run all unit test scenarios"""
    with codraft_app_context() as win:
        test_signal_features(win)
        test_image_features(win)


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Z-log scale test

Testing z-log scale tool feature.
"""

import numpy as np
from guiqwt.builder import make

from codraft import patch  # pylint: disable=unused-import
from codraft.tests.data import create_2d_steps_data
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.vistools import view_image_items

SHOW = True  # Show test in GUI-based test launcher


def zlogscale_test():
    """Z-log scale test"""
    with qt_app_context():
        data = create_2d_steps_data(1024, width=256, dtype=np.int32)
        item = make.image(data)
        view_image_items([item], title="Z-log scale test")
        item.set_zaxis_log_state(True)  # pylint: disable=no-member


if __name__ == "__main__":
    zlogscale_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT
=======

CodraFT is a generic signal and image processing software based on Python scientific
libraries (such as NumPy, SciPy or scikit-image) and Qt graphical user interfaces
(thanks to guidata and guiqwt libraries).

CodraFT is Copyright © 2022 CEA-CODRA, Pierre Raybaut, and Licensed under the
terms of the BSD 3-Clause License or the CeCILL-B License.
"""

from distutils.core import setup

setup()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Image peak detection test: testing algorithm limits
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...


from codraft.core.computation.image import get_2d_peaks_coords
from codraft.env import execenv
from codraft.tests.data import get_peak2d_data
from codraft.tests.peak2d_unit import exec_image_peak_detection_func
from codraft.utils.qthelpers import qt_app_context

SHOW = False  # Do not show test in GUI-based test launcher


def peak2d_limit_test():
    """2D peak detection test"""
    with qt_app_context():
        execenv.print("Testing peak detection algorithm with random generated data:")
        for idx in range(100):
            execenv.print(f"  Iteration #{idx:02d}: ", end="")
            generated_data = get_peak2d_data(multi=True)
            coords = get_2d_peaks_coords(generated_data)
            if coords.shape[0] != 4:
                execenv.print(f"KO - {coords.shape[0]}/4 peaks were detected")
                exec_image_peak_detection_func(generated_data)
            else:
                execenv.print("OK")
        # Showing results for last generated sample
        exec_image_peak_detection_func(generated_data)


if __name__ == "__main__":
    peak2d_limit_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Contour finding application test
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from codraft.core.gui.processor.image import ContourShapeParam
from codraft.tests import codraft_app_context, take_plotwidget_screenshot
from codraft.tests.data import create_test_image3
from codraft.tests.roi_app import create_test_image_with_roi

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run ROI unit test scenario"""
    size = 200
    with codraft_app_context() as win:
        panel = win.imagepanel
        ima1 = create_test_image3(size)
        panel.add_object(ima1)
        param = ContourShapeParam()
        panel.processor.compute_contour_shape(param)
        take_plotwidget_screenshot(panel, "contour_test")
        ima2 = create_test_image_with_roi(size)
        panel.add_object(ima2)
        panel.processor.edit_regions_of_interest()
        panel.processor.compute_contour_shape(param)


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Basic application launcher test 1

Running application a few times in a row with different entry parameters.
"""

from codraft import app
from codraft.env import execenv
from codraft.utils.qthelpers import qt_app_context

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Testing CodraFT app launcher"""
    with qt_app_context(exec_loop=True):
        execenv.print("Opening CodraFT with no argument")
        app.create()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Flat field test

Testing the following:
  - Create a gaussian image (raw data)
  - Create a random noised image (flat data)
  - Compute the flat field image
"""

from codraft.config import _
from codraft.core.gui.processor.image import FlatFieldParam
from codraft.core.model.base import UniformRandomParam
from codraft.core.model.image import Gauss2DParam, ImageTypes, new_image_param
from codraft.tests import codraft_app_context

SHOW = True  # Show test in GUI-based test launcher


def test():
    """Run flat field test scenario"""
    with codraft_app_context() as win:
        panel = win.imagepanel

        i0p = new_image_param(_("Raw data (2D-Gaussian)"), itype=ImageTypes.GAUSS)
        panel.new_object(i0p, addparam=Gauss2DParam(), edit=False)
        i1p = new_image_param(
            _("Flat data (Uniform random)"), itype=ImageTypes.UNIFORMRANDOM
        )
        addp = UniformRandomParam()
        addp.vmax = 5
        panel.new_object(i1p, addparam=addp, edit=False)

        panel.objlist.select_rows((0, 1))
        ffp = FlatFieldParam()
        ffp.threshold = 80
        panel.processor.flat_field_correction(ffp)


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Datasets
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

from copy import deepcopy

import guidata.dataset.dataitems as gdi
import guidata.dataset.datatypes as gdt
import numpy as np
from guidata.configtools import get_icon
from guidata.utils import update_dataset
from guiqwt.builder import make
from guiqwt.curve import CurveItem
from guiqwt.styles import update_style_attr

from codraft.config import Conf, _
from codraft.core.computation import fit
from codraft.core.model import base
from codraft.env import execenv


class SignalParam(gdt.DataSet, base.ObjectItf):
    """Signal dataset"""

    CONF_FMT = Conf.view.sig_format
    DEFAULT_FMT = ".3f"
    VALID_DTYPES = (np.float32, np.float64, np.complex128)

    _tabs = gdt.BeginTabGroup("all")

    _datag = gdt.BeginGroup(_("Data and metadata"))
    title = gdi.StringItem(_("Signal title"), default=_("Untitled"))
    xydata = gdi.FloatArrayItem(_("Data"), transpose=True, minmax="rows")
    metadata = base.MetadataItem(_("Metadata"), default={})
    _e_datag = gdt.EndGroup(_("Data and metadata"))

    _unitsg = gdt.BeginGroup(_("Titles and units"))
    title = gdi.StringItem(_("Signal title"), default=_("Untitled"))
    _tabs_u = gdt.BeginTabGroup("units")
    _unitsx = gdt.BeginGroup(_("X-axis"))
    xlabel = gdi.StringItem(_("Title"), default="")
    xunit = gdi.StringItem(_("Unit"), default="")
    _e_unitsx = gdt.EndGroup(_("X-axis"))
    _unitsy = gdt.BeginGroup(_("Y-axis"))
    ylabel = gdi.StringItem(_("Title"), default="")
    yunit = gdi.StringItem(_("Unit"), default="")
    _e_unitsy = gdt.EndGroup(_("Y-axis"))
    _e_tabs_u = gdt.EndTabGroup("units")
    _e_unitsg = gdt.EndGroup(_("Titles and units"))

    _e_tabs = gdt.EndTabGroup("all")

    def copy_data_from(self, other, dtype=None):
        """Copy data from other dataset instance"""
        if dtype not in (None, float, complex, np.complex128):
            raise RuntimeError("Signal data only supports float64/complex128 dtype")
        self.metadata = deepcopy(other.metadata)
        self.xydata = np.array(other.xydata, copy=True, dtype=dtype)

    def set_data_type(self, dtype):  # pylint: disable=unused-argument,no-self-use
        """Change data type"""
        raise RuntimeError("Setting data type is not support for signals")

    def set_xydata(self, x, y, dx=None, dy=None):
        """Set xy data"""
        if x is not None:
            x = np.array(x)
        if y is not None:
            y = np.array(y)
        if dx is not None:
            dx = np.array(dx)
        if dy is not None:
            dy = np.array(dy)
        if dx is None and dy is None:
            self.xydata = np.vstack([x, y])
        else:
            if dx is None:
                dx = np.zeros_like(dy)
            else:
                dy = np.zeros_like(dx)
            self.xydata = np.vstack((x, y, dx, dy))

    def __get_x(self):
        """Get x data"""
        if self.xydata is not None:
            return self.xydata[0]
        return None

    def __set_x(self, data):
        """Set x data"""
        self.xydata[0] = np.array(data)

    def __get_y(self):
        """Get y data"""
        if self.xydata is not None:
            return self.xydata[1]
        return None

    def __set_y(self, data):
        """Set y data"""
        self.xydata[1] = np.array(data)

    x = property(__get_x, __set_x)
    y = data = property(__get_y, __set_y)

    def get_data(self, roi_index: int = None) -> np.ndarray:
        """
        Return original data (if ROI is not defined or `roi_index` is None),
        or ROI data (if both ROI and `roi_index` are defined).
        """
        if self.roi is None or roi_index is None:
            return self.x, self.y
        i1, i2 = self.roi[roi_index, :]
        return self.x[i1:i2], self.y[i1:i2]

    def make_item(self, update_from=None):
        """Make plot item from data"""
        if len(self.xydata) == 2:  # x, y signal
            x, y = self.xydata
            item = make.mcurve(x.real, y.real, label=self.title)
        elif len(self.xydata) == 3:  # x, y, dy error bar signal
            x, y, dy = self.xydata
            item = make.merror(x.real, y.real, dy.real, label=self.title)
        elif len(self.xydata) == 4:  # x, y, dx, dy error bar signal
            x, y, dx, dy = self.xydata
            item = make.merror(x.real, y.real, dx.real, dy.real, label=self.title)
        else:
            raise RuntimeError("data not supported")
        if update_from is not None:
            update_dataset(item.curveparam, update_from.curveparam)
        return item

    def update_item(self, item: CurveItem, ref_item: CurveItem = None):
        """Update plot item from data"""
        if len(self.xydata) == 2:  # x, y signal
            x, y = self.xydata
            item.set_data(x.real, y.real)
        elif len(self.xydata) == 3:  # x, y, dy error bar signal
            x, y, dy = self.xydata
            item.set_data(x.real, y.real, dy=dy.real)
        elif len(self.xydata) == 4:  # x, y, dx, dy error bar signal
            x, y, dx, dy = self.xydata
            item.set_data(x.real, y.real, dx.real, dy.real)
        item.curveparam.label = self.title
        if execenv.demo_mode:
            item.curveparam.line.width = 3
        update_style_attr(next(make.style), item.curveparam)
        update_dataset(item.curveparam, self.metadata)
        item.update_params()

    def roi_coords_to_indexes(self, coords: list) -> np.ndarray:
        """Convert ROI coordinates to indexes"""
        indexes = np.array(coords, int)
        for row in range(indexes.shape[0]):
            for col in range(indexes.shape[1]):
                x0 = coords[row][col]
                indexes[row, col] = np.abs(self.x - x0).argmin()
        return indexes

    def get_roi_param(self, title, *defaults):
        """Return ROI parameters dataset"""
        imax = len(self.x) - 1
        i0, i1 = defaults

        class ROIParam(gdt.DataSet):
            """Signal ROI parameters"""

            col1 = gdi.IntItem(_("First point index"), default=i0, min=-1, max=imax)
            col2 = gdi.IntItem(_("Last point index"), default=i1, min=-1, max=imax)

        return ROIParam(title)

    @staticmethod
    def params_to_roidata(params: gdt.DataSetGroup) -> np.ndarray:
        """Convert list of dataset parameters to ROI data"""
        roilist = []
        for roiparam in params.datasets:
            roilist.append([roiparam.col1, roiparam.col2])
        if len(roilist) == 0:
            return None
        return np.array(roilist, int)

    def new_roi_item(self, fmt, lbl, editable):
        """Return a new ROI item from scratch"""
        coords = self.x.min(), self.x.max()
        return base.make_roi_item(
            lambda x, y, _title: make.range(x, y), coords, "ROI", fmt, lbl, editable
        )

    def iterate_roi_items(self, fmt: str, lbl: bool, editable: bool = True):
        """Make plot item representing a Region of Interest"""
        if self.roi is not None:
            for index, coords in enumerate(self.x[self.roi]):
                yield base.make_roi_item(
                    lambda x, y, _title: make.range(x, y),
                    coords,
                    f"ROI{index:02d}",
                    fmt,
                    lbl,
                    editable,
                )


def create_signal(
    title: str,
    x: np.ndarray = None,
    y: np.ndarray = None,
    dx: np.ndarray = None,
    dy: np.ndarray = None,
    metadata: dict = None,
    units: tuple = None,
    labels: tuple = None,
) -> SignalParam:
    """Create a new Signal object

    :param str title: signal title
    :param numpy.ndarray x: X data
    :param numpy.ndarray y: Y data
    :param numpy.ndarray dx: dX data (optional: error bars)
    :param numpy.ndarray dy: dY data (optional: error bars)
    :param dict metadata: signal metadata
    :param tuple units: X, Y units (tuple of strings)
    :param tuple labels: X, Y labels (tuple of strings)
    """
    assert isinstance(title, str)
    signal = SignalParam()
    signal.title = title
    signal.set_xydata(x, y, dx=dx, dy=dy)
    if units is not None:
        signal.xunit, signal.yunit = units
    if labels is not None:
        signal.xlabel, signal.ylabel = labels
    signal.metadata = {} if metadata is None else metadata
    return signal


class SignalTypes(base.Choices):
    """Signal types"""

    ZEROS = _("zeros")
    GAUSS = _("gaussian")
    LORENTZ = _("lorentzian")
    VOIGT = "Voigt"
    UNIFORMRANDOM = _("random (uniform law)")
    NORMALRANDOM = _("random (normal law)")


class GaussLorentzVoigtParam(gdt.DataSet):
    """Parameters for Gaussian and Lorentzian functions"""

    a = gdi.FloatItem("A", default=1.0)
    ymin = gdi.FloatItem("Ymin", default=0.0).set_pos(col=1)
    sigma = gdi.FloatItem("σ", default=1.0)
    mu = gdi.FloatItem("μ", default=0.0).set_pos(col=1)


class SignalParamNew(gdt.DataSet):
    """New signal dataset"""

    title = gdi.StringItem(_("Title"), default=_("Untitled"))
    xmin = gdi.FloatItem("Xmin", default=-10.0)
    xmax = gdi.FloatItem("Xmax", default=10.0)
    size = gdi.IntItem(
        _("Size"), help=_("Signal size (total number of points)"), min=1, default=500
    )
    type = gdi.ChoiceItem(_("Type"), SignalTypes.get_choices())


def new_signal_param(title=None, stype=None, xmin=None, xmax=None, size=None):
    """Create a new Signal dataset instance.

    :param str title: dataset title (default: None, uses default title)"""
    if title is None:
        title = _("Untitled signal")
    param = SignalParamNew(title=title, icon=get_icon("new_signal.svg"))
    param.title = title
    if xmin is not None:
        param.xmin = xmin
    if xmax is not None:
        param.xmax = xmax
    if size is not None:
        param.size = size
    if stype is not None:
        param.type = stype
    return param


SIG_NB = 0


def create_signal_from_param(newparam, addparam=None, edit=False, parent=None):
    """Create a new Signal object from a dialog box.

    :param SignalParamNew param: new signal parameters
    :param guidata.dataset.datatypes.DataSet addparam: additional parameters
    :param bool edit: Open a dialog box to edit parameters (default: False)
    :param QWidget parent: parent widget
    """
    global SIG_NB  # pylint: disable=global-statement
    if newparam is None:
        newparam = new_signal_param()
    incr_sig_nb = not newparam.title
    if incr_sig_nb:
        newparam.title = f"{newparam.title} {SIG_NB + 1:d}"
    if not edit or addparam is not None or newparam.edit(parent=parent):
        if incr_sig_nb:
            SIG_NB += 1
        signal = create_signal(newparam.title)
        xarr = np.linspace(newparam.xmin, newparam.xmax, newparam.size)
        p = addparam
        if newparam.type == SignalTypes.ZEROS:
            signal.set_xydata(xarr, np.zeros(newparam.size))
        elif newparam.type in (SignalTypes.UNIFORMRANDOM, SignalTypes.NORMALRANDOM):
            pclass = {
                SignalTypes.UNIFORMRANDOM: base.UniformRandomParam,
                SignalTypes.NORMALRANDOM: base.NormalRandomParam,
            }[newparam.type]
            if p is None:
                p = pclass(_("Signal") + " - " + newparam.type.value)
            if edit and not p.edit(parent=parent):
                return None
            rng = np.random.default_rng(p.seed)
            if newparam.type == SignalTypes.UNIFORMRANDOM:
                yarr = rng.random((newparam.size,)) * (p.vmax - p.vmin) + p.vmin
            elif newparam.type == SignalTypes.NORMALRANDOM:
                yarr = rng.normal(p.mu, p.sigma, size=(newparam.size,))
            else:
                raise NotImplementedError(f"New param type: {newparam.type.value}")
            signal.set_xydata(xarr, yarr)
        elif newparam.type == SignalTypes.GAUSS:
            if p is None:
                p = GaussLorentzVoigtParam(_("New gaussian function"))
            if edit and not p.edit(parent=parent):
                return None
            yarr = fit.GaussianModel.func(xarr, p.a, p.sigma, p.mu, p.ymin)
            signal.set_xydata(xarr, yarr)
        elif newparam.type == SignalTypes.LORENTZ:
            if p is None:
                p = GaussLorentzVoigtParam(_("New lorentzian function"))
            if edit and not p.edit(parent=parent):
                return None
            yarr = fit.LorentzianModel.func(xarr, p.a, p.sigma, p.mu, p.ymin)
            signal.set_xydata(xarr, yarr)
        elif newparam.type == SignalTypes.VOIGT:
            if p is None:
                p = GaussLorentzVoigtParam(_("New Voigt function"))
            if edit and not p.edit(parent=parent):
                return None
            yarr = fit.VoigtModel.func(xarr, p.a, p.sigma, p.mu, p.ymin)
            signal.set_xydata(xarr, yarr)
        return signal
    return None



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Dependencies viewer test
"""

from codraft.utils.qthelpers import qt_app_context
from codraft.widgets.instconfviewer import exec_codraft_installconfig_dialog

SHOW = True  # Show test in GUI-based test launcher


def test_dep_viewer():
    """Test dep viewer window"""
    with qt_app_context():
        exec_codraft_installconfig_dialog()


if __name__ == "__main__":
    test_dep_viewer()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
Image centroid computation test

Comparing different algorithms for centroid calculation:

- SciPy (measurements.center_of_mass)
- OpenCV (moments)
- Method based on moments
- Method based on Fourier (CodraFT)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

import time

import numpy as np
import scipy.ndimage as spi
from guiqwt.builder import make
from numpy import ma

from codraft.config import _
from codraft.core.computation.image import get_centroid_fourier
from codraft.env import execenv
from codraft.tests.data import get_laser_spot_data
from codraft.utils.qthelpers import qt_app_context
from codraft.utils.vistools import view_image_items

SHOW = True  # Show test in GUI-based test launcher


def get_centroid_from_moments(data):
    """Computing centroid from image moments"""
    y, x = np.ogrid[: data.shape[0], : data.shape[1]]
    imx, imy = data.sum(axis=0)[None, :], data.sum(axis=1)[:, None]
    m00 = np.array(data, dtype=float).sum() or 1.0
    m10 = (np.array(imx, dtype=float) * x).sum() / m00
    m01 = (np.array(imy, dtype=float) * y).sum() / m00
    return int(m01), int(m10)


def get_centroid_with_cv2(data):
    """Compute centroid from moments with OpenCV"""
    import cv2  # pylint: disable=import-outside-toplevel

    m = cv2.moments(data)
    col = int(m["m10"] / m["m00"])
    row = int(m["m01"] / m["m00"])
    return row, col


def add_xcursor(items, x, y, title):
    """Added X cursor to plot"""
    label = "  " + f'{_("Centroid")}[{title}] (x=%s, y=%s)'
    execenv.print(label % (x, y))
    cursor = make.xcursor(x, y, label=label)
    cursor.setTitle(title)
    items.append(cursor)


def compare_centroid_funcs(data):
    """Compare centroid methods"""
    items = []
    items += [make.image(data, interpolation="nearest", eliminate_outliers=2.0)]
    # Computing centroid coordinates
    for name, func in (
        ("SciPy", spi.center_of_mass),
        ("OpenCV", get_centroid_with_cv2),
        ("Moments", get_centroid_from_moments),
        ("Fourier", get_centroid_fourier),
    ):
        try:
            t0 = time.time()
            row, col = func(data)
            dt = time.time() - t0
            add_xcursor(items, col, row, name)
            execenv.print(f"    Calculation time: {int(dt * 1e3):d} ms")
        except ImportError:
            execenv.print(f"    Unable to compute {name}: missing module")
    view_image_items(items)


def centroid_test():
    """Centroid test"""
    with qt_app_context():
        for data in get_laser_spot_data():
            execenv.print(f"Data[dtype={data.dtype},shape={data.shape}]")
            # Testing with masked arrays
            compare_centroid_funcs(data.view(ma.MaskedArray))


if __name__ == "__main__":
    centroid_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Signal Processor GUI module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

import re

import numpy as np
import scipy.integrate as spt
import scipy.ndimage as spi
import scipy.optimize as spo
import scipy.signal as sps
from guidata.dataset.dataitems import ChoiceItem, FloatItem, IntItem
from guidata.dataset.datatypes import DataSet, DataSetGroup

from codraft.config import _
from codraft.core.computation import fit
from codraft.core.computation.signal import (
    derivative,
    moving_average,
    normalize,
    peak_indexes,
    xpeak,
    xy_fft,
    xy_ifft,
)
from codraft.core.gui.processor.base import BaseProcessor, ClipParam, ThresholdParam
from codraft.core.model.base import ResultShape, ShapeTypes
from codraft.core.model.signal import SignalParam, create_signal
from codraft.utils.qthelpers import exec_dialog, qt_try_except
from codraft.widgets import fitdialog, signalpeakdialog


class PeakDetectionParam(DataSet):
    """Peak detection parameters"""

    threshold = IntItem(
        _("Threshold"), default=30, min=0, max=100, slider=True, unit="%"
    )
    min_dist = IntItem(_("Minimum distance"), default=1, min=1, unit="points")


class NormalizeParam(DataSet):
    """Normalize parameters"""

    methods = (
        (_("maximum"), "maximum"),
        (_("amplitude"), "amplitude"),
        (_("sum"), "sum"),
        (_("energy"), "energy"),
    )
    method = ChoiceItem(_("Normalize with respect to"), methods)


class XYCalibrateParam(DataSet):
    """Signal calibration parameters"""

    axes = (("x", _("X-axis")), ("y", _("Y-axis")))
    axis = ChoiceItem(_("Calibrate"), axes, default="y")
    a = FloatItem("a", default=1.0)
    b = FloatItem("b", default=0.0)


class PolynomialFitParam(DataSet):
    """Polynomial fitting parameters"""

    degree = IntItem(_("Degree"), 3, min=1, max=10, slider=True)


class FWHMParam(DataSet):
    """FWHM parameters"""

    fittypes = (
        ("GaussianModel", _("Gaussian")),
        ("LorentzianModel", _("Lorentzian")),
        ("VoigtModel", "Voigt"),
    )

    fittype = ChoiceItem(_("Fit type"), fittypes, default="GaussianModel")


class SignalProcessor(BaseProcessor):
    """Object handling signal processing: operations, processing, computing"""

    # pylint: disable=duplicate-code

    def extract_roi(self, roidata: np.ndarray = None, singleobj: bool = None) -> None:
        """Extract Region Of Interest (ROI) from data"""
        roieditordata = self._get_roieditordata(roidata, singleobj)
        if roieditordata is None or roieditordata.is_empty:
            return
        obj = self.objlist.get_sel_object()
        group = obj.roidata_to_params(roieditordata.roidata)

        if roieditordata.singleobj:

            def suffix_func(group: DataSetGroup):
                if len(group.datasets) == 1:
                    p = group.datasets[0]
                    return f"indexes={p.col1:d}:{p.col2:d}"
                return ""

            def extract_roi_func(x, y, group: DataSetGroup):
                """Extract ROI function"""
                xout, yout = np.ones_like(x) * np.nan, np.ones_like(y) * np.nan
                for p in group.datasets:
                    slice0 = slice(p.col1, p.col2 + 1)
                    xout[slice0], yout[slice0] = x[slice0], y[slice0]
                nans = np.isnan(xout) | np.isnan(yout)
                return xout[~nans], yout[~nans]

            # TODO: [P2] Instead of removing geometric shapes, apply roi extract
            self.compute_11(
                "ROI",
                extract_roi_func,
                group,
                suffix=suffix_func,
                func_obj=lambda obj, _group: obj.remove_resultshapes(),
                edit=False,
            )
        else:
            # TODO: [P2] Instead of removing geometric shapes, apply roi extract
            self.compute_1n(
                [f"ROI{iroi}" for iroi in range(len(group.datasets))],
                lambda x, y, p: (x[p.col1 : p.col2 + 1], y[p.col1 : p.col2 + 1]),
                group.datasets,
                suffix=lambda p: f"indexes={p.col1:d}:{p.col2:d}",
                func_obj=lambda obj, _group: obj.remove_resultshapes(),
                edit=False,
            )

    def swap_axes(self):
        """Swap data axes"""
        self.compute_11(
            "SwapAxes",
            lambda x, y: (y, x),
            func_obj=lambda obj: obj.remove_resultshapes(),
        )

    def compute_abs(self):
        """Compute absolute value"""
        self.compute_11("Abs", lambda x, y: (x, np.abs(y)))

    def compute_log10(self):
        """Compute Log10"""
        self.compute_11("Log10", lambda x, y: (x, np.log10(y)))

    def detect_peaks(self, param: PeakDetectionParam = None) -> None:
        """Detect peaks from data"""
        obj = self.objlist.get_sel_object()
        edit = param is None
        if edit:
            dlg = signalpeakdialog.SignalPeakDetectionDialog(self.panel)
            dlg.setup_data(obj.x, obj.y)
            if exec_dialog(dlg):
                param = PeakDetectionParam(_("Peak detection"))
                param.threshold = int(dlg.get_threshold() * 100)
                param.min_dist = dlg.get_min_dist()

        def func(x, y, p):
            """Peak detection"""
            indexes = peak_indexes(y, thres=p.threshold * 0.01, min_dist=p.min_dist)
            return x[indexes], y[indexes]

        def func_obj(obj, param):  # pylint: disable=unused-argument
            """Customize signal object"""
            obj.metadata["curvestyle"] = "Sticks"

        self.compute_11(
            "Peaks",
            func,
            param,
            suffix=lambda p: f"threshold={p.threshold}%, min_dist={p.min_dist}pts",
            func_obj=func_obj,
            edit=edit,
        )

    # ------Signal Processing
    def apply_11_func(self, obj, orig, func, param, message):
        """Apply 11 function: 1 object in --> 1 object out"""

        # (self is used by @qt_try_except)
        # pylint: disable=unused-argument
        @qt_try_except(message)
        def apply_11_func_callback(self, obj, orig, func, param):
            """Apply 11 function callback: 1 object in --> 1 object out"""
            data = orig.xydata
            if len(data) == 2:  # x, y signal
                x, y = data
                if param is None:
                    obj.xydata = func(x, y)
                else:
                    obj.xydata = func(x, y, param)
            elif len(data) == 4:  # x, y, dx, dy error bar signal
                x, y, dx, dy = data
                if param is None:
                    x2, y2 = func(x, y)
                    _x3, dy2 = func(x, dy)
                else:
                    x2, y2 = func(x, y, param)
                    _x3, dy2 = func(x, dy, param)
                obj.xydata = x2, y2, dx, dy2

        return apply_11_func_callback(self, obj, orig, func, param)

    @qt_try_except()
    def normalize(self, param: NormalizeParam = None) -> None:
        """Normalize data"""
        edit = param is None
        if edit:
            param = NormalizeParam(_("Normalize"))

        def func(x, y, p):
            return (x, normalize(y, p.method))

        self.compute_11(
            "Normalize", func, param, suffix=lambda p: f"ref={p.method}", edit=edit
        )

    @qt_try_except()
    def compute_derivative(self):
        """Compute derivative"""
        self.compute_11("Derivative", lambda x, y: (x, derivative(x, y)))

    @qt_try_except()
    def compute_integral(self):
        """Compute integral"""
        self.compute_11("Integral", lambda x, y: (x, spt.cumtrapz(y, x, initial=0.0)))

    @qt_try_except()
    def calibrate(self, param: XYCalibrateParam = None) -> None:
        """Compute data linear calibration"""
        edit = param is None
        if edit:
            param = XYCalibrateParam(_("Linear calibration"), "y = a.x + b")

        def func(x, y, p):
            """Compute linear calibration"""
            if p.axis == "x":
                return p.a * x + p.b, y
            return x, p.a * y + p.b

        self.compute_11(
            "LinearCal",
            func,
            param,
            suffix=lambda p: f"{p.axis}={p.a}*{p.axis}+{p.b}",
            edit=edit,
        )

    @qt_try_except()
    def compute_threshold(self, param: ThresholdParam = None) -> None:
        """Compute threshold clipping"""
        edit = param is None
        if edit:
            param = ThresholdParam(_("Thresholding"))
        self.compute_11(
            "Threshold",
            lambda x, y, p: (x, np.clip(y, p.value, y.max())),
            param,
            suffix=lambda p: f"min={p.value} lsb",
            edit=edit,
        )

    @qt_try_except()
    def compute_clip(self, param: ClipParam = None) -> None:
        """Compute maximum data clipping"""
        edit = param is None
        if edit:
            param = ClipParam(_("Clipping"))
        self.compute_11(
            "Clip",
            lambda x, y, p: (x, np.clip(y, y.min(), p.value)),
            param,
            suffix=lambda p: f"max={p.value} lsb",
            edit=edit,
        )

    @staticmethod
    def func_gaussian_filter(x, y, p):
        """Compute gaussian filter"""
        return (x, spi.gaussian_filter1d(y, p.sigma))

    @staticmethod
    def func_moving_average(x, y, p):
        """Moving average computing function"""
        return (x, moving_average(y, p.n))

    @staticmethod
    def func_moving_median(x, y, p):
        """Moving median computing function"""
        return (x, sps.medfilt(y, kernel_size=p.n))

    @qt_try_except()
    def compute_wiener(self):
        """Compute Wiener filter"""
        self.compute_11("WienerFilter", lambda x, y: (x, sps.wiener(y)))

    @qt_try_except()
    def compute_fft(self):
        """Compute iFFT"""
        self.compute_11("FFT", xy_fft)

    @qt_try_except()
    def compute_ifft(self):
        """Compute FFT"""
        self.compute_11("iFFT", xy_ifft)

    @qt_try_except()
    def compute_fit(self, name, fitdlgfunc):
        """Compute fitting curve"""
        rows = self.objlist.get_selected_rows()
        for row in rows:
            self.__row_compute_fit(row, name, fitdlgfunc)

    @qt_try_except()
    def compute_polyfit(self, param: PolynomialFitParam = None) -> None:
        """Compute polynomial fitting curve"""
        txt = _("Polynomial fit")
        edit = param is None
        if edit:
            param = PolynomialFitParam(txt)
        if not edit or param.edit(self):
            dlgfunc = fitdialog.polynomialfit
            self.compute_fit(
                txt,
                lambda x, y, degree=param.degree, parent=self.panel.parent(): dlgfunc(
                    x, y, degree, parent=parent
                ),
            )

    def __row_compute_fit(self, row, name, fitdlgfunc):
        """Curve fitting computing sub-method"""
        obj = self.objlist[row]
        output = fitdlgfunc(obj.x, obj.y, parent=self.panel.parent())
        if output is not None:
            y, params = output
            results = {}
            for param in params:
                if re.match(r"[\S\_]*\d{2}$", param.name):
                    shname = param.name[:-2]
                    value = results.get(shname, np.array([]))
                    results[shname] = np.array(list(value) + [param.value])
                else:
                    results[param.name] = param.value
            # Creating new signal
            signal = create_signal(f"{name}({obj.title})", obj.x, y, metadata=results)
            # Creating new plot item
            self.panel.add_object(signal, refresh=False)
            # Refreshing list
            self.objlist.refresh_list(-1)

    @qt_try_except()
    def compute_multigaussianfit(self):
        """Compute multi-Gaussian fitting curve"""
        rows = self.objlist.get_selected_rows()
        fitdlgfunc = fitdialog.multigaussianfit
        for row in rows:
            dlg = signalpeakdialog.SignalPeakDetectionDialog(self.panel)
            obj = self.objlist[row]
            dlg.setup_data(obj.x, obj.y)
            if exec_dialog(dlg):
                # Computing x, y
                peaks = dlg.get_peak_indexes()
                self.__row_compute_fit(
                    row,
                    _("Multi-Gaussian fit"),
                    lambda x, y, peaks=peaks, parent=self.panel.parent(): fitdlgfunc(
                        x, y, peaks, parent=parent
                    ),
                )

    # ------Signal Computing
    def apply_10_func(self, orig, func, param, message) -> ResultShape:
        """Apply 10 function: 1 object in --> 0 object out (scalar result)"""

        # (self is used by @qt_try_except)
        # pylint: disable=unused-argument
        @qt_try_except(message)
        def apply_10_func_callback(self, orig, func, param):
            """Apply 10 function cb: 1 object in --> 0 object out (scalar result)"""
            if param is None:
                return func(orig)
            return func(orig, param)

        return apply_10_func_callback(self, orig, func, param)

    @qt_try_except()
    def compute_fwhm(self, param: FWHMParam = None) -> None:
        """Compute FWHM"""
        title = _("FWHM")

        def fwhm(signal: SignalParam, param: FWHMParam):
            """Compute FWHM"""
            res = []
            for i_roi in signal.iterate_roi_indexes():
                x, y = signal.get_data(i_roi)
                dx = np.max(x) - np.min(x)
                dy = np.max(y) - np.min(y)
                base = np.min(y)
                sigma, mu = dx * 0.1, xpeak(x, y)
                FitModel = getattr(fit, param.fittype)
                amp = FitModel.get_amp_from_amplitude(dy, sigma)

                def func(params):
                    """Fitting model function"""
                    # pylint: disable=cell-var-from-loop
                    return y - FitModel.func(x, *params)

                (amp, sigma, mu, base), _ier = spo.leastsq(
                    func, np.array([amp, sigma, mu, base])
                )
                x0, y0, x1, y1 = FitModel.half_max_segment(amp, sigma, mu, base)
                res.append([i_roi, x0, y0, x1, y1])
            return signal.add_resultshape(title, ShapeTypes.SEGMENT, np.array(res))

        edit = param is None
        if edit:
            param = FWHMParam(title)
        self.compute_10(title, fwhm, param, edit=edit)

    @qt_try_except()
    def compute_fw1e2(self):
        """Compute FW at 1/e²"""
        title = _("FW") + "1/e²"

        def fw1e2(signal: SignalParam):
            """Compute FW at 1/e²"""
            res = []
            for i_roi in signal.iterate_roi_indexes():
                x, y = signal.get_data(i_roi)
                dx = np.max(x) - np.min(x)
                dy = np.max(y) - np.min(y)
                base = np.min(y)
                sigma, mu = dx * 0.1, xpeak(x, y)
                amp = fit.GaussianModel.get_amp_from_amplitude(dy, sigma)
                p_in = np.array([amp, sigma, mu, base])

                def func(params):
                    """Fitting model function"""
                    # pylint: disable=cell-var-from-loop
                    return y - fit.GaussianModel.func(x, *params)

                p_out, _ier = spo.leastsq(func, p_in)
                amp, sigma, mu, base = p_out
                hw = 2 * sigma
                amplitude = fit.GaussianModel.amplitude(amp, sigma)
                yhm = amplitude / np.e**2 + base
                res.append([i_roi, mu - hw, yhm, mu + hw, yhm])
            return signal.add_resultshape(title, ShapeTypes.SEGMENT, np.array(res))

        self.compute_10(title, fw1e2)

    def _get_stat_funcs(self):
        """Return statistics functions list"""
        return [
            ("min(y)", lambda xy: xy[1].min()),
            ("max(y)", lambda xy: xy[1].max()),
            ("<y>", lambda xy: xy[1].mean()),
            ("Median(y)", lambda xy: np.median(xy[1])),
            ("σ(y)", lambda xy: xy[1].std()),
            ("Σ(y)", lambda xy: xy[1].sum()),
            ("∫ydx", lambda xy: np.trapz(xy[1], xy[0])),
        ]



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT Dockable widgets
"""

from guidata.qthelpers import is_dark_mode
from guidata.qtwidgets import DockableWidget, DockableWidgetMixin
from guiqwt.plot import ImageWidget
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW


class DockablePlotWidget(DockableWidget):
    """Docked plotting widget"""

    LOCATION = QC.Qt.RightDockWidgetArea

    def __init__(self, parent, plotwidgetclass, toolbar):
        super().__init__(parent)
        self.toolbar = toolbar
        layout = QW.QVBoxLayout()
        self.plotwidget = plotwidgetclass()
        layout.addWidget(self.plotwidget)
        self.setLayout(layout)
        self.setup()

    def get_plot(self):
        """Return plot instance"""
        return self.plotwidget.plot

    def setup(self):
        """Setup plotting widget"""
        title = self.toolbar.windowTitle()
        pwidget = self.plotwidget
        pwidget.add_toolbar(self.toolbar, title)
        if isinstance(self.plotwidget, ImageWidget):
            pwidget.register_all_image_tools()
        else:
            pwidget.register_all_curve_tools()
        #  Customizing widget appearances
        plot = pwidget.get_plot()
        if not is_dark_mode():
            for widget in (pwidget, plot, self):
                widget.setBackgroundRole(QG.QPalette.Window)
                widget.setAutoFillBackground(True)
                widget.setPalette(QG.QPalette(QC.Qt.white))
        canvas = plot.canvas()
        canvas.setFrameStyle(canvas.Plain | canvas.NoFrame)

    # ------DockableWidget API
    def visibility_changed(self, enable):
        """DockWidget visibility has changed"""
        DockableWidget.visibility_changed(self, enable)
        self.toolbar.setVisible(enable)


class DockableTabWidget(QW.QTabWidget, DockableWidgetMixin):
    """Docked tab widget"""

    LOCATION = QC.Qt.LeftDockWidgetArea



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""Curve fitting dialog widgets"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...


import numpy as np
from guidata.configtools import get_icon
from guiqwt.widgets.fit import FitDialog, FitParam

from codraft.config import _
from codraft.core.computation import fit
from codraft.core.computation.signal import xpeak
from codraft.utils.qthelpers import exec_dialog
from codraft.utils.tests import get_default_test_name


def guifit(
    x,
    y,
    fitfunc,
    fitparams,
    fitargs=None,
    fitkwargs=None,
    wintitle=None,
    title=None,
    xlabel=None,
    ylabel=None,
    param_cols=1,
    auto_fit=True,
    winsize=None,
    winpos=None,
    parent=None,
    name=None,
):
    """GUI-based curve fitting tool"""
    win = FitDialog(
        edit=True,
        wintitle=wintitle,
        icon=None,
        toolbar=True,
        options=dict(title=title, xlabel=xlabel, ylabel=ylabel),
        parent=parent,
        param_cols=param_cols,
        auto_fit=auto_fit,
    )
    if name is None:
        name = get_default_test_name()
    win.setObjectName(name)
    win.set_data(x, y, fitfunc, fitparams, fitargs, fitkwargs)
    win.autofit()  # TODO: [P3] make this optional
    if parent is None:
        win.setWindowIcon(get_icon("codraft.svg"))
    if winsize is not None:
        win.resize(*winsize)
    if winpos is not None:
        win.move(*winpos)
    win.get_plot().do_autoscale()
    if exec_dialog(win):
        return win.get_values()
    return None


# --- Polynomial fitting curve -------------------------------------------------
def polynomialfit(x, y, degree, parent=None, name=None):
    """Compute polynomial fit

    Returns (yfit, params), where yfit is the fitted curve and params are
    the fitting parameters"""
    ivals = np.polyfit(x, y, degree)

    params = []
    for index in range(degree + 1):
        val = ivals[index]
        vmax = max(1.0, np.abs(val))
        param = FitParam(f"c{(len(ivals) - index - 1):d}", val, -2 * vmax, 2 * vmax)
        params.append(param)

    def fitfunc(x, params):
        return np.polyval(params, x)

    values = guifit(
        x, y, fitfunc, params, parent=parent, wintitle=_("Polymomial fit"), name=name
    )
    if values:
        return fitfunc(x, values), params


# --- Gaussian fitting curve ---------------------------------------------------
def gaussianfit(x, y, parent=None, name=None):
    """Compute Gaussian fit

    Returns (yfit, params), where yfit is the fitted curve and params are
    the fitting parameters"""
    dx = np.max(x) - np.min(x)
    dy = np.max(y) - np.min(y)
    sigma = dx * 0.1
    amp = fit.GaussianModel.get_amp_from_amplitude(dy, sigma)

    a = FitParam(_("Amplitude"), amp, 0.0, amp * 1.2)
    b = FitParam(_("Base line"), np.min(y), np.min(y) - 0.1 * dy, np.max(y))
    sigma = FitParam(_("Std-dev") + " (σ)", sigma, sigma * 0.2, sigma * 10)
    mu = FitParam(_("Mean") + " (μ)", xpeak(x, y), np.min(x), np.max(x))

    params = [a, sigma, mu, b]

    def fitfunc(x, params):
        return fit.GaussianModel.func(x, *params)

    values = guifit(
        x, y, fitfunc, params, parent=parent, wintitle=_("Gaussian fit"), name=name
    )
    if values:
        return fitfunc(x, values), params


# --- Lorentzian fitting curve -------------------------------------------------
def lorentzianfit(x, y, parent=None, name=None):
    """Compute Lorentzian fit

    Returns (yfit, params), where yfit is the fitted curve and params are
    the fitting parameters"""
    dx = np.max(x) - np.min(x)
    dy = np.max(y) - np.min(y)
    sigma = dx * 0.1
    amp = fit.LorentzianModel.get_amp_from_amplitude(dy, sigma)

    a = FitParam(_("Amplitude"), amp, 0.0, amp * 1.2)
    b = FitParam(_("Base line"), np.min(y), np.min(y) - 0.1 * dy, np.max(y))
    sigma = FitParam(_("Std-dev") + " (σ)", sigma, sigma * 0.2, sigma * 10)
    mu = FitParam(_("Mean") + " (μ)", xpeak(x, y), np.min(x), np.max(x))

    params = [a, sigma, mu, b]

    def fitfunc(x, params):
        return fit.LorentzianModel.func(x, *params)

    values = guifit(
        x, y, fitfunc, params, parent=parent, wintitle=_("Lorentzian fit"), name=name
    )
    if values:
        return fitfunc(x, values), params


# --- Voigt fitting curve ------------------------------------------------------
def voigtfit(x, y, parent=None, name=None):
    """Compute Voigt fit

    Returns (yfit, params), where yfit is the fitted curve and params are
    the fitting parameters"""
    dx = np.max(x) - np.min(x)
    dy = np.max(y) - np.min(y)
    sigma = dx * 0.1
    amp = fit.VoigtModel.get_amp_from_amplitude(dy, sigma)

    a = FitParam(_("Amplitude"), amp, 0.0, amp * 1.2)
    b = FitParam(_("Base line"), np.min(y), np.min(y) - 0.1 * dy, np.max(y))
    sigma = FitParam(_("Std-dev") + " (σ)", sigma, sigma * 0.2, sigma * 10)
    mu = FitParam(_("Mean") + " (μ)", xpeak(x, y), np.min(x), np.max(x))

    params = [a, sigma, mu, b]

    def fitfunc(x, params):
        return fit.VoigtModel.func(x, *params)

    values = guifit(
        x, y, fitfunc, params, parent=parent, wintitle=_("Voigt fit"), name=name
    )
    if values:
        return fitfunc(x, values), params


# --- Multi-Gaussian fitting curve ---------------------------------------------
def multigaussian(x, *values, **kwargs):
    """Return a 1-dimensional multi-Gaussian function."""
    a_amp = values[0::2]
    a_sigma = values[1::2]
    y0 = values[-1]
    a_x0 = kwargs["a_x0"]
    y = np.zeros_like(x) + y0
    for amp, sigma, x0 in zip(a_amp, a_sigma, a_x0):
        y += amp * np.exp(-0.5 * ((x - x0) / sigma) ** 2)
    return y


def multigaussianfit(x, y, peak_indexes, parent=None, name=None):
    """Compute Multi-Gaussian fit

    Returns (yfit, params), where yfit is the fitted curve and params are
    the fitting parameters"""
    params = []
    for index, i0 in enumerate(peak_indexes):
        istart = 0
        iend = len(x) - 1
        if index > 0:
            istart = (peak_indexes[index - 1] + i0) // 2
        if index < len(peak_indexes) - 1:
            iend = (peak_indexes[index + 1] + i0) // 2
        dx = 0.5 * (x[iend] - x[istart])
        dy = np.max(y[istart:iend]) - np.min(y[istart:iend])
        stri = f"{index + 1:02d}"
        params += [
            FitParam(("A") + stri, y[i0], 0.0, dy * 2),
            FitParam("σ" + stri, dx / 10, dx / 100, dx),
        ]

    params.append(
        FitParam(
            _("Y0"), np.min(y), np.min(y) - 0.1 * (np.max(y) - np.min(y)), np.max(y)
        )
    )

    kwargs = dict(a_x0=x[peak_indexes])

    def fitfunc(xi, params):
        return multigaussian(xi, *params, **kwargs)

    param_cols = 1
    if len(params) > 8:
        param_cols = 4
    values = guifit(
        x,
        y,
        fitfunc,
        params,
        param_cols=param_cols,
        winsize=(900, 600),
        parent=parent,
        name=name,
        wintitle=_("Multi-Gaussian fit"),
    )
    if values:
        return fitfunc(x, values), params



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/__init__.py for details)

"""
CodraFT unit tests
"""

from contextlib import contextmanager

from guidata.guitest import run_testlauncher
from qtpy import QtCore as QC

import codraft.config  # Loading icons
from codraft import env
from codraft.core.gui.main import CodraFTMainWindow
from codraft.utils import qthelpers as qth
from codraft.utils import tests

# TODO: [P2] Documentation: add more screenshots from tests
# TODO: [P3] Create subpackages "app" & "unit" + add support for subpackages in
# test launcher


@contextmanager
def codraft_app_context(size=None, maximized=False, save=False, console=None):
    """Context manager handling CodraFT mainwindow creation and Qt event loop"""
    if size is None:
        size = 950, 600

    with qth.qt_app_context(exec_loop=True):
        try:
            win = CodraFTMainWindow(console=console)
            if maximized:
                win.showMaximized()
            else:
                width, height = size
                win.resize(width, height)
                win.showNormal()
            win.show()
            win.setObjectName(tests.get_default_test_name())  # screenshot name
            yield win
        finally:
            if save:
                try:
                    win.save_to_h5_file(tests.get_output_data_path("h5"))
                except PermissionError:
                    pass


def take_plotwidget_screenshot(panel, name):
    """Eventually takes plotwidget screenshot (only in screenshot mode)"""
    if env.execenv.screenshot:
        qth.grab_save_window(panel.itmlist.plotwidget, f"{panel.PREFIX}_{name}")


def run():
    """Run CodraFT test launcher"""
    run_testlauncher(codraft)


if __name__ == "__main__":
    run()
